"""Test utils."""
import numpy as np
from scipy.sparse import (csr_matrix, csc_matrix, isspmatrix,
                          bsr_matrix, isspmatrix_bsr,
                          spdiags)

from numpy.testing import (TestCase, assert_equal, assert_almost_equal,
                           assert_array_almost_equal, assert_array_equal)

import pyamg

from pyamg.util.utils import (diag_sparse, profile_solver, to_type,
                              type_prep, get_diagonal,
                              scale_rows, scale_columns,
                              get_block_diag, symmetric_rescaling, symmetric_rescaling_sa,
                              filter_operator, scale_T,
                              get_Cpt_params, compute_BtBinv, eliminate_diag_dom_nodes)

from pyamg.relaxation.utils import relaxation_as_linear_operator


class TestUtils(TestCase):
    def test_diag_sparse(self):
        # check sparse -> array
        A = np.array([[-4]])
        assert_equal(diag_sparse(csr_matrix(A)), [-4])

        A = np.array([[1, 0, -5], [-2, 5, 0]])
        assert_equal(diag_sparse(csr_matrix(A)), [1, 5])

        A = np.array([[0, 1], [0, -5]])
        assert_equal(diag_sparse(csr_matrix(A)), [0, -5])

        A = np.array([[1.3, -4.7, 0], [-2.23, 5.5, 0], [9, 0, -2]])
        assert_equal(diag_sparse(csr_matrix(A)), [1.3, 5.5, -2])

        # check array -> sparse
        A = np.array([[-4]])
        assert_equal(diag_sparse(np.array([-4])).toarray(),
                     csr_matrix(A).toarray())

        A = np.array([[1, 0], [0, 5]])
        assert_equal(diag_sparse(np.array([1, 5])).toarray(),
                     csr_matrix(A).toarray())

        A = np.array([[0, 0], [0, -5]])
        assert_equal(diag_sparse(np.array([0, -5])).toarray(),
                     csr_matrix(A).toarray())

        A = np.array([[1.3, 0, 0], [0, 5.5, 0], [0, 0, -2]])
        assert_equal(diag_sparse(np.array([1.3, 5.5, -2])).toarray(),
                     csr_matrix(A).toarray())

    def test_scale_rows(self):
        Aorig = np.array([[0.0, 1, 0], [2, 0, 3], [4, 5, 6]])
        v = np.array([7.0, 8, 9])
        Ascaled = np.array([[0.0, 7, 0], [16, 0, 24], [36, 45, 54]])

        # test CSR
        A = csr_matrix(Aorig)
        A = scale_rows(A, v)
        assert_equal(A.toarray(), Ascaled)

        # test CSC
        A = csc_matrix(Aorig)
        A = scale_rows(A, v)
        assert_equal(A.toarray(), Ascaled)

    def test_scale_columns(self):
        Aorig = np.array([[0.0, 1, 0], [2, 0, 3], [4, 5, 6]]).T
        v = np.array([7.0, 8, 9])
        Ascaled = np.array([[0.0, 7, 0], [16, 0, 24], [36, 45, 54]]).T

        # test CSR
        A = csr_matrix(Aorig)
        A = scale_columns(A, v)
        assert_equal(A.toarray(), Ascaled)

        # test CSC
        A = csc_matrix(Aorig)
        A = scale_columns(A, v)
        assert_equal(A.toarray(), Ascaled)

    def test_symmetric_rescaling(self):
        cases = []
        cases.append(diag_sparse(np.array([1, 2, 3, 4])))
        cases.append(diag_sparse(np.array([1, 0, 3, 4])))

        A = np.array([[5.5, 3.5, 4.8],
                      [2., 9.9, 0.5],
                      [6.5, 2.6, 5.7]])

        # test csr
        A = csr_matrix(A)
        cases.append(A)
        P = diag_sparse([1, 0, 1])
        cases.append(P*A*P)
        P = diag_sparse([0, 1, 0])
        cases.append(P*A*P)
        P = diag_sparse([1, -1, 1])
        cases.append(P*A*P)

        # test csc
        A = csc_matrix(A)
        cases.append(A)

        for A in cases:
            D_sqrt, D_sqrt_inv, DAD = symmetric_rescaling(A)

            assert_almost_equal(diag_sparse(A) > 0, diag_sparse(DAD))
            assert_almost_equal(diag_sparse(DAD), D_sqrt*D_sqrt_inv)

            D_sqrt, D_sqrt_inv = diag_sparse(D_sqrt), diag_sparse(D_sqrt_inv)
            assert_almost_equal((D_sqrt_inv*A*D_sqrt_inv).toarray(),
                                DAD.toarray())

    def test_symmetric_rescaling_sa(self):
        cases = []
        # case 1
        e = np.ones((5, 1)).ravel()
        data = [-1*e, 2*e, -1*e]
        A = spdiags(data, [-1, 0, 1], 5, 5).tocsr()
        B = e.copy().reshape(-1, 1)
        DAD_answer = np.array([[1., -0.5, 0., 0., 0.],
                               [-0.5, 1., -0.5, 0., 0.],
                               [0., -0.5, 1., -0.5, 0.],
                               [0., 0., -0.5, 1., -0.5],
                               [0., 0., 0., -0.5, 1.]])
        DB_answer = np.sqrt(2*e.reshape(-1, 1))
        #           matrix   B    BH   expected matrix  expected B  expected BH
        cases.append((A, B, None, DAD_answer, DB_answer, None))

        # case 2
        A2 = A.copy()
        A2.symmetry = 'nonsymmetric'
        cases.append((A2, B.copy(), B.copy(), DAD_answer, DB_answer,
                      DB_answer))

        # case 3
        A3 = A.copy()
        A3.symmetry = 'hermitian'
        cases.append((A3, B.copy(), None, DAD_answer, DB_answer, None))

        # case 4
        B4 = np.hstack((B.copy(), 2*B.copy()))
        DB4_answer = np.sqrt(2)*B4
        A4 = A.copy()
        A4.symmetry = 'nonsymmetric'
        cases.append((A4, B4, B4.copy(), DAD_answer, DB4_answer, DB4_answer))

        for case in cases:
            [A, B, BH, DAD_answer, DB_answer, DBH_answer] = case

            [DAD, DB, DBH] = symmetric_rescaling_sa(A, B, BH=BH)
            assert_array_almost_equal(DAD.toarray(), DAD_answer)
            assert_array_almost_equal(DB, DB_answer)
            if DBH_answer is not None:
                assert_array_almost_equal(DBH, DBH_answer)

    def test_profile_solver(self):
        from scipy.sparse.linalg import cg
        from pyamg.gallery import poisson
        from pyamg.aggregation import smoothed_aggregation_solver

        A = poisson((100, 100), format='csr')
        ml = smoothed_aggregation_solver(A)

        opts = []
        opts.append({})
        opts.append({'accel': cg, 'atol': 'legacy'})
        opts.append({'accel': cg, 'tol': 1e-10, 'atol': 'legacy'})

        for kwargs in opts:
            residuals = profile_solver(ml, **kwargs)
        del residuals

    def test_get_block_diag(self):
        A = csr_matrix(np.arange(1, 37, dtype=float).reshape(6, 6))
        block_diag = get_block_diag(A, blocksize=1, inv_flag=False)
        assert_array_almost_equal(np.ravel(block_diag), A.diagonal())

        block_diag = get_block_diag(A, blocksize=2, inv_flag=False)
        answer = np.array([[[1., 2.],
                            [7., 8.]],
                           [[15., 16.],
                            [21., 22.]],
                           [[29., 30.],
                            [35., 36.]]])
        assert_array_almost_equal(np.ravel(block_diag), np.ravel(answer))

        block_diag_inv = get_block_diag(A, blocksize=2, inv_flag=True)
        answer = np.array([[[-1.33333333, 0.33333333],
                            [1.16666667, -0.16666667]],
                           [[-3.66666667, 2.66666667],
                            [3.5, -2.5]],
                           [[-6., 5.],
                            [5.83333333, -4.83333333]]])
        assert_array_almost_equal(np.ravel(block_diag_inv), np.ravel(answer),
                                  decimal=3)

        # try singular (1,1) block, a zero (2,2) block and a zero (0,2) block
        A = bsr_matrix(np.array([1., 2., 3., 4., 0., 0.,
                                 5., 6., 7., 8., 0., 0.,
                                 9., 10., 11., 11., 12., 13.,
                                 14., 15., 16., 16., 18., 19.,
                                 20., 21., 22., 23., 0., 0.,
                                 26., 27., 28., 29., 0., 0.]).reshape(6, 6),
                       blocksize=(3, 3))
        block_diag_inv = get_block_diag(A, blocksize=2, inv_flag=True)
        answer = np.array([[[-1.5, 0.5],
                            [1.25, -0.25]],
                           [[0.01458886, 0.02122016],
                            [0.01458886, 0.02122016]],
                           [[0., 0.],
                            [0., 0.]]])
        assert_array_almost_equal(np.ravel(block_diag_inv), np.ravel(answer),
                                  decimal=3)

        # try with different types of zero blocks
        A = bsr_matrix(np.array([0., 0., 3., 4., 0., 0.,
                                 0., 0., 7., 8., 0., 0.,
                                 0., 0., 0., 0., 0., 0.,
                                 0., 0., 0., 0., 0., 0.,
                                 0., 0., 0., 0., 22., 23.,
                                 0., 0., 0., 0., 28., 29.]).reshape(6, 6),
                       blocksize=(2, 2))
        block_diag_inv = get_block_diag(A, blocksize=2, inv_flag=False)
        answer = np.array([[[0., 0.],
                            [0., 0.]],
                           [[0., 0.],
                            [0., 0.]],
                           [[22., 23.],
                            [28., 29.]]])
        assert_array_almost_equal(np.ravel(block_diag_inv), np.ravel(answer),
                                  decimal=3)

    def test_relaxation_as_linear_operator(self):
        As = []
        bs = []
        xs = []
        methods = ['gauss_seidel', 'jacobi', 'block_gauss_seidel',
                   'block_jacobi']
        params = [{}, {'iterations': 2}]
        As.append(pyamg.gallery.poisson((10, 10), format='csr'))
        As.append(1.0j*pyamg.gallery.poisson((10, 10), format='csr'))
        As.append(1.0j*pyamg.gallery.elasticity.linear_elasticity((20, 20))[0])
        As.append(pyamg.gallery.elasticity.linear_elasticity((20, 20))[0])
        for A in As:
            if A.dtype == 'complex':
                xs.append(np.random.rand(A.shape[0], 1)
                          + 1.0j * np.random.rand(A.shape[0], 1))
                bs.append(np.random.rand(A.shape[0], 1)
                          + 1.0j * np.random.rand(A.shape[0], 1))
            else:
                bs.append(np.random.rand(A.shape[0], 1))
                xs.append(np.random.rand(A.shape[0], 1))

        for method in methods:
            for kwargs in params:
                for (A, x, b) in zip(As, xs, bs):
                    kwargs_linop = dict(kwargs)
                    # run relaxation as a linear operator
                    if kwargs_linop == dict({}):
                        relax = relaxation_as_linear_operator(method, A, b)
                    else:
                        fmethod = (method, kwargs_linop)
                        relax = relaxation_as_linear_operator(fmethod, A, b)
                    x_linop = relax * x

                    # manually run the relaxation routine
                    relax2 = getattr(pyamg.relaxation.relaxation, method)
                    x_gold = x.copy()
                    blockflag = False
                    kwargs_gold = dict(kwargs)
                    # deal with block matrices
                    if method.startswith('block') and isspmatrix_bsr(A):
                        blockflag = True
                        kwargs_gold['blocksize'] = A.blocksize[0]
                    # deal with omega and jacobi
                    # --> note that we assume the default setup for jacobi uses
                    # omega = 1/rho
                    if method.endswith('jacobi'):
                        if blockflag:
                            kwargs_gold['omega'] = 1.0/A.rho_block_D_inv
                        else:
                            kwargs_gold['omega'] = 1.0/A.rho_D_inv

                    relax2(A, x_gold, b, **kwargs_gold)

                    assert_array_almost_equal(x_linop, x_gold)

    def test_filter_operator(self):
        # Basic tests of dimension 1 and 2 problems
        # 1x1
        A = csr_matrix(np.array([[1.2]]))
        C = csr_matrix(np.array([[1.]]))
        B = np.array([[0.5]])
        Bf = np.array([[1.5]])
        A_filter = filter_operator(A, C, B, Bf).toarray()
        A_known = np.array([[3.0]])
        assert_array_almost_equal(A_known, A_filter)
        # 1x1, but with no entries in C
        C = csr_matrix(np.array([[0.]]))
        A_filter = filter_operator(A, C, B, Bf).toarray()
        A_known = np.array([[0.0]])
        assert_array_almost_equal(A_known, A_filter)
        # 1x1, but with no entries in A
        A = csr_matrix(np.array([[0.]]))
        C = csr_matrix(np.array([[1.]]))
        A_filter = filter_operator(A, C, B, Bf).toarray()
        A_known = np.array([[3.0]])
        assert_array_almost_equal(A_known, A_filter)

        # 1x2
        A = csr_matrix(np.array([[1.2, 1.]]))
        C = csr_matrix(np.array([[1., 1.]]))
        B = np.array([[0.5], [0.5]])
        Bf = np.array([[1.5]])
        A_filter = filter_operator(A, C, B, Bf).toarray()
        A_known = np.array([[1.6, 1.4]])
        assert_array_almost_equal(A_known, A_filter)
        # 1x2, but sparser
        C = csr_matrix(np.array([[0., 1.]]))
        A_filter = filter_operator(A, C, B, Bf).toarray()
        A_known = np.array([[0., 3.]])
        assert_array_almost_equal(A_known, A_filter)
        # 1x2, but with no entries
        C = csr_matrix(np.array([[0., 0.]]))
        A_filter = filter_operator(A, C, B, Bf).toarray()
        A_known = np.array([[0., 0.]])
        assert_array_almost_equal(A_known, A_filter)

        # 2x1
        A = csr_matrix(np.array([[1.2], [1.]]))
        C = csr_matrix(np.array([[1.], [1.]]))
        B = np.array([[0.5]])
        Bf = np.array([[1.5], [0.4]])
        A_filter = filter_operator(A, C, B, Bf).toarray()
        A_known = np.array([[3.], [0.8]])
        assert_array_almost_equal(A_known, A_filter)
        # 2x1, but sparser
        C = csr_matrix(np.array([[0.], [1.]]))
        A_filter = filter_operator(A, C, B, Bf).toarray()
        A_known = np.array([[0.], [.8]])
        assert_array_almost_equal(A_known, A_filter)
        # 2x1, but with no entries
        C = csr_matrix(np.array([[0.], [0.]]))
        A_filter = filter_operator(A, C, B, Bf).toarray()
        A_known = np.array([[0.], [0.]])
        assert_array_almost_equal(A_known, A_filter)

        # 2x2
        A = csr_matrix(np.array([[1.2, 1.1], [1., 0.5]]))
        C = csr_matrix(np.array([[1.2, 1.1], [1., 0.]]))
        B = np.array([[0.5, 1.0], [0.5, 1.1]])
        Bf = np.array([[0.5, 1.0], [0.5, 1.1]])
        A_filter = filter_operator(A, C, B, Bf).toarray()
        A_known = np.array([[1., 0.], [1.08, 0.]])
        assert_array_almost_equal(A_known, A_filter)
        # 1x2, but sparser
        C = csr_matrix(np.array([[0., 0.], [1., 0.]]))
        A_filter = filter_operator(A, C, B, Bf).toarray()
        A_known = np.array([[0., 0.], [1.08, 0.]])
        assert_array_almost_equal(A_known, A_filter)
        # Try block structure
        A = A.tobsr((2, 2))
        C = C.tobsr((2, 2))
        A_filter = filter_operator(A, C, B, Bf).toarray()
        A_known = np.array([[1., 0.], [0., 1.]])
        assert_array_almost_equal(A_known, A_filter)

        # Basic tests, with easy to compute answers
        # test one, the constant
        A = np.array([[1., 1, 1],
                      [1, 1, 1],
                      [0, 1, 0],
                      [0, 1, 0],
                      [0, 0, 1],
                      [0, 0, 1]])
        C = np.array([[1., 1, 0],
                      [1, 1, 0],
                      [0, 1, 0],
                      [0, 1, 0],
                      [0, 0, 1],
                      [0, 0, 1]])
        B = np.ones((3, 1))
        Bf = np.ones((6, 1))
        A_filter = filter_operator(csr_matrix(A), csr_matrix(C), B, Bf)
        A_filter = A_filter.toarray()
        A_known = np.array([[0.5, 0.5, 0.],
                            [0.5, 0.5, 0.],
                            [0., 1., 0.],
                            [0., 1., 0.],
                            [0., 0., 1.],
                            [0., 0., 1.]])
        assert_array_almost_equal(A_known, A_filter)
        # test two, the constant and linears
        B = np.hstack((B, np.arange(B.shape[0]).reshape(-1, 1)))
        Bf = np.hstack((Bf, np.arange(Bf.shape[0]).reshape(-1, 1)))
        A_filter = filter_operator(csr_matrix(A), csr_matrix(C), B, Bf)
        A_filter = A_filter.toarray()
        A_known = np.array([[1., 0., 0.],
                            [0., 1., 0.],
                            [0., 1.5, 0.],
                            [0., 2., 0.],
                            [0., 0., 1.8],
                            [0., 0., 2.2]])
        assert_array_almost_equal(A_known, A_filter)

        # Run two tests based on the Laplacian
        # first test, constants
        from pyamg.gallery import poisson
        A = poisson((10, 10), format='csr')
        C = A.copy()
        C.data[np.arange(0, C.nnz, 5)] = 0.0
        C.eliminate_zeros()
        B = np.ones((A.shape[0], 1))
        Bf = np.ones((A.shape[0], 1))
        A_filter = filter_operator(A, C, B, Bf)
        assert_array_almost_equal(A_filter*B, Bf)
        # second test, constants and linears
        B = np.hstack((B, np.arange(B.shape[0]).reshape(-1, 1)))
        Bf = np.hstack((Bf, np.arange(Bf.shape[0]).reshape(-1, 1)))
        A_filter = filter_operator(A, C, B, Bf)
        assert_array_almost_equal(A_filter*B, Bf)

    def test_scale_T(self):
        from scipy.sparse import bsr_matrix

        # Trivially sized tests
        # 1x1
        T = np.array([[1.1]])
        P_I = np.array([[1.0]])
        I_F = np.array([[0.0]])
        T_scaled = scale_T(bsr_matrix(T), bsr_matrix(P_I), bsr_matrix(I_F))
        T_scaled = T_scaled.toarray()
        T_answer = np.array([[1.0]])
        assert_array_almost_equal(T_answer, T_scaled)

        T = np.array([[1.1]])
        P_I = np.array([[0.0]])
        I_F = np.array([[1.0]])
        T_scaled = scale_T(bsr_matrix(T), bsr_matrix(P_I), bsr_matrix(I_F))
        T_scaled = T_scaled.toarray()
        T_answer = np.array([[1.1]])
        assert_array_almost_equal(T_answer, T_scaled)

        T = np.array([[0.0]])
        P_I = np.array([[0.0]])
        I_F = np.array([[1.0]])
        T_scaled = scale_T(bsr_matrix(T), bsr_matrix(P_I), bsr_matrix(I_F))
        T_scaled = T_scaled.toarray()
        T_answer = np.array([[0.]])
        assert_array_almost_equal(T_answer, T_scaled)

        # 2x1
        T = np.array([[1.5], [1.2]])
        P_I = np.array([[1.], [0.]])
        I_F = np.array([[0., 0.], [0., 1.]])
        T_scaled = scale_T(bsr_matrix(T), bsr_matrix(P_I), bsr_matrix(I_F))
        T_scaled = T_scaled.toarray()
        T_answer = np.array([[1.], [0.8]])
        assert_array_almost_equal(T_answer, T_scaled)

        T = np.array([[0.], [1.2]])
        P_I = np.array([[1.], [0.]])
        I_F = np.array([[0., 0.], [0., 1.]])
        T_scaled = scale_T(bsr_matrix(T), bsr_matrix(P_I), bsr_matrix(I_F))
        T_scaled = T_scaled.toarray()
        T_answer = np.array([[1.], [0.]])
        assert_array_almost_equal(T_answer, T_scaled)

        T = np.array([[0.], [0.]])
        P_I = np.array([[1.], [0.]])
        I_F = np.array([[0., 0.], [0., 1.]])
        T_scaled = scale_T(bsr_matrix(T), bsr_matrix(P_I), bsr_matrix(I_F))
        T_scaled = T_scaled.toarray()
        T_answer = np.array([[1.], [0.]])
        assert_array_almost_equal(T_answer, T_scaled)

        T = np.array([[0.], [0.]])
        P_I = np.array([[0.], [0.]])
        I_F = np.array([[1., 0.], [0., 1.]])
        T_scaled = scale_T(bsr_matrix(T), bsr_matrix(P_I), bsr_matrix(I_F))
        T_scaled = T_scaled.toarray()
        T_answer = np.array([[0.], [0.]])
        assert_array_almost_equal(T_answer, T_scaled)

        # 2x2
        T = np.array([[2., 0.], [1., 1.]])
        P_I = np.array([[1., 0.], [0., 1.]])
        I_F = np.array([[0., 0.], [0., 0.]])
        T_scaled = scale_T(bsr_matrix(T, blocksize=(1, 1)),
                           bsr_matrix(P_I, blocksize=(1, 1)),
                           bsr_matrix(I_F, blocksize=(1, 1))).toarray()
        T_answer = np.array([[1., 0.], [0., 1.]])
        assert_array_almost_equal(T_answer, T_scaled)

        T = np.array([[2., 0.], [1., 1.]])
        P_I = np.array([[1., 0.], [0., 1.]])
        I_F = np.array([[0., 0.], [0., 0.]])
        T_scaled = scale_T(bsr_matrix(T, blocksize=(2, 2)),
                           bsr_matrix(P_I, blocksize=(2, 2)),
                           bsr_matrix(I_F, blocksize=(2, 2))).toarray()
        T_answer = np.array([[1., 0.], [0., 1.]])
        assert_array_almost_equal(T_answer, T_scaled)

        T = np.array([[2., 0.], [1., 1.]])
        P_I = np.array([[0., 0.], [0., 0.]])
        I_F = np.array([[1., 0.], [0., 1.]])
        T_scaled = scale_T(bsr_matrix(T, blocksize=(2, 2)),
                           bsr_matrix(P_I, blocksize=(2, 2)),
                           bsr_matrix(I_F, blocksize=(2, 2))).toarray()
        T_answer = np.array([[2., 0.], [1., 1.]])
        assert_array_almost_equal(T_answer, T_scaled)

        # Test for one CSR and one BSR example
        T = np.array([[1.0, 0., 0.],
                      [0.5, 0., 0.],
                      [0., 1., 0.],
                      [0., 0.5, 0.],
                      [0., 0., 1.],
                      [0., 0., 0.25]])
        P_I = np.array([[0., 0., 0.],
                        [1., 0., 0.],
                        [0., 1., 0.],
                        [0., 0., 0.],
                        [0., 0., 0.],
                        [0., 0., 1.]])
        I_F = np.array([[1., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 1., 0., 0.],
                        [0., 0., 0., 0., 1., 0.],
                        [0., 0., 0., 0., 0., 0.]])
        T_answer = np.array([[2., 0., 0.],
                             [1., 0., 0.],
                             [0., 1., 0.],
                             [0., 0.5, 0.],
                             [0., 0., 4.],
                             [0., 0., 1.]])
        T_scaled = scale_T(bsr_matrix(T), bsr_matrix(P_I), bsr_matrix(I_F))
        T_scaled = T_scaled.toarray()
        assert_array_almost_equal(T_answer, T_scaled)

        # BSR test
        T = np.array([[1.0, 1., 0., 0.],
                      [0.5, 1., 0., 0.],
                      [1., 0., 0., 0.],
                      [0., 1., 0., 0.],
                      [0., 0., 2., 1.],
                      [0., 0., 3., 1.],
                      [0., 0., 4., 1.],
                      [0., 0., 2., 0.]])
        P_I = np.array([[0., 0., 0., 0.],
                        [0., 0., 0., 0.],
                        [1., 0., 0., 0.],
                        [0., 1., 0., 0.],
                        [0., 0., 1., 0.],
                        [0., 0., 0., 1.],
                        [0., 0., 0., 0.],
                        [0., 0., 0., 0.]])
        I_F = np.array([[1., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 1., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 1., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 1.]])
        T_answer = np.array([[1., 1., 0., 0.],
                             [0.5, 1., 0., 0.],
                             [1., 0., 0., 0.],
                             [0., 1., 0., 0.],
                             [0., 0., 1., 0.],
                             [0., 0., 0., 1.],
                             [0., 0., -1., 2.],
                             [0., 0., -2., 2.]])
        T = bsr_matrix(T, blocksize=(2, 2))
        P_I = bsr_matrix(P_I, blocksize=(2, 2))
        I_F = bsr_matrix(I_F, blocksize=(2, 2))
        T_scaled = scale_T(T, P_I, I_F).toarray()
        assert_array_almost_equal(T_answer, T_scaled)

        # BSR test
        T = np.array([[1.0, 1., 0., 0.],
                      [0.5, 1., 0., 0.],
                      [1., 1., 0., 0.],
                      [1., 1., 0., 0.],
                      [0., 0., 2., 1.],
                      [0., 0., 3., 1.],
                      [0., 0., 4., 1.],
                      [0., 0., 2., 0.]])
        P_I = np.array([[0., 0., 0., 0.],
                        [0., 0., 0., 0.],
                        [1., 0., 0., 0.],
                        [0., 1., 0., 0.],
                        [0., 0., 1., 0.],
                        [0., 0., 0., 1.],
                        [0., 0., 0., 0.],
                        [0., 0., 0., 0.]])
        I_F = np.array([[1., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 1., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 1., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 1.]])
        T_answer = np.array([[0.5, 0.5, 0., 0.],
                             [0.375, 0.375, 0., 0.],
                             [1., 0., 0., 0.],
                             [0., 1., 0., 0.],
                             [0., 0., 1., 0.],
                             [0., 0., 0., 1.],
                             [0., 0., -1., 2.],
                             [0., 0., -2., 2.]])
        # Cpts = np.array([1, 2])
        T = bsr_matrix(T, blocksize=(2, 2))
        P_I = bsr_matrix(P_I, blocksize=(2, 2))
        I_F = bsr_matrix(I_F, blocksize=(2, 2))
        T_scaled = scale_T(T, P_I, I_F).toarray()
        assert_array_almost_equal(T_answer, T_scaled)

    def test_get_Cpt_params(self):
        from pyamg.gallery import poisson
        from scipy.sparse import csr_matrix, bsr_matrix

        # Begin with trivially sized tests
        # 1x1
        A = csr_matrix(np.array([[1.2]]))
        Cpts = np.array([0])
        AggOp = csr_matrix(np.array([[1.]]))
        T = AggOp.copy().tobsr()
        params = get_Cpt_params(A, Cpts, AggOp, T)
        I_C = bsr_matrix(np.array([[1.]]), blocksize=(1, 1))
        I_F = bsr_matrix(np.array([[0.]]), blocksize=(1, 1))
        P_I = bsr_matrix(np.array([[1.]]), blocksize=(1, 1))
        assert_equal(np.array([0]), params['Cpts'])
        assert_equal(np.array([]), params['Fpts'])
        assert_equal(I_C.indptr, params['I_C'].indptr)
        assert_equal(I_C.indices, params['I_C'].indices)
        assert_equal(I_C.data, params['I_C'].data)
        assert_equal(I_F.indptr, params['I_F'].indptr)
        assert_equal(I_F.indices, params['I_F'].indices)
        assert_equal(I_F.data, params['I_F'].data)
        assert_equal(P_I.indptr, params['P_I'].indptr)
        assert_equal(P_I.indices, params['P_I'].indices)
        assert_equal(P_I.data, params['P_I'].data)

        A = csr_matrix(np.array([[1.2]]))
        Cpts = np.array([])
        AggOp = csr_matrix(np.array([[1.]]))
        T = AggOp.copy().tobsr()
        params = get_Cpt_params(A, Cpts, AggOp, T)
        I_C = bsr_matrix(np.array([[0.]]), blocksize=(1, 1))
        I_F = bsr_matrix(np.array([[1.]]), blocksize=(1, 1))
        P_I = bsr_matrix(np.array([[0.]]), blocksize=(1, 1))
        assert_equal(np.array([]), params['Cpts'])
        assert_equal(np.array([0]), params['Fpts'])
        assert_equal(I_C.indptr, params['I_C'].indptr)
        assert_equal(I_C.indices, params['I_C'].indices)
        assert_equal(I_C.data, params['I_C'].data)
        assert_equal(I_F.indptr, params['I_F'].indptr)
        assert_equal(I_F.indices, params['I_F'].indices)
        assert_equal(I_F.data, params['I_F'].data)
        assert_equal(P_I.indptr, params['P_I'].indptr)
        assert_equal(P_I.indices, params['P_I'].indices)
        assert_equal(P_I.data, params['P_I'].data)

        # 2x2
        A = csr_matrix(np.array([[1., 1.], [1., 1.]]))
        Cpts = np.array([0])
        AggOp = csr_matrix(np.array([[1.], [1.]]))
        T = AggOp.copy().tobsr()
        params = get_Cpt_params(A, Cpts, AggOp, T)
        I_C = bsr_matrix(np.array([[1., 0.], [0., 0.]]), blocksize=(1, 1))
        I_F = bsr_matrix(np.array([[0., 0.], [0., 1.]]), blocksize=(1, 1))
        P_I = bsr_matrix(np.array([[1.], [0.]]), blocksize=(1, 1))
        assert_equal(np.array([0]), params['Cpts'])
        assert_equal(np.array([1]), params['Fpts'])
        assert_equal(I_C.indptr, params['I_C'].indptr)
        assert_equal(I_C.indices, params['I_C'].indices)
        assert_equal(I_C.data, params['I_C'].data)
        assert_equal(I_F.indptr, params['I_F'].indptr)
        assert_equal(I_F.indices, params['I_F'].indices)
        assert_equal(I_F.data, params['I_F'].data)
        assert_equal(P_I.indptr, params['P_I'].indptr)
        assert_equal(P_I.indices, params['P_I'].indices)
        assert_equal(P_I.data, params['P_I'].data)

        Cpts = np.array([0, 1])
        AggOp = csr_matrix(np.array([[1., 0], [0., 1.]]))
        T = AggOp.copy().tobsr()
        params = get_Cpt_params(A, Cpts, AggOp, T)
        I_C = bsr_matrix(np.array([[1., 0.], [0., 1.]]), blocksize=(1, 1))
        I_F = bsr_matrix(np.array([[0., 0.], [0., 0.]]), blocksize=(1, 1))
        P_I = bsr_matrix(np.array([[1., 0.], [0., 1.]]), blocksize=(1, 1))
        assert_equal(np.array([0, 1]), params['Cpts'])
        assert_equal(np.array([]), params['Fpts'])
        assert_equal(I_C.indptr, params['I_C'].indptr)
        assert_equal(I_C.indices, params['I_C'].indices)
        assert_equal(I_C.data, params['I_C'].data)
        assert_equal(I_F.indptr, params['I_F'].indptr)
        assert_equal(I_F.indices, params['I_F'].indices)
        assert_equal(I_F.data, params['I_F'].data)
        assert_equal(P_I.indptr, params['P_I'].indptr)
        assert_equal(P_I.indices, params['P_I'].indices)
        assert_equal(P_I.data, params['P_I'].data)

        Cpts = np.array([])
        AggOp = csr_matrix(np.array([[0.], [0.]]))
        T = AggOp.copy().tobsr()
        params = get_Cpt_params(A, Cpts, AggOp, T)
        I_C = bsr_matrix(np.array([[0., 0.], [0., 0.]]), blocksize=(1, 1))
        I_F = bsr_matrix(np.array([[1., 0.], [0., 1.]]), blocksize=(1, 1))
        P_I = bsr_matrix(np.array([[0.], [0.]]), blocksize=(1, 1))
        assert_equal(np.array([]), params['Cpts'])
        assert_equal(np.array([0, 1]), params['Fpts'])
        assert_equal(I_C.indptr, params['I_C'].indptr)
        assert_equal(I_C.indices, params['I_C'].indices)
        assert_equal(I_C.data, params['I_C'].data)
        assert_equal(I_F.indptr, params['I_F'].indptr)
        assert_equal(I_F.indices, params['I_F'].indices)
        assert_equal(I_F.data, params['I_F'].data)
        assert_equal(P_I.indptr, params['P_I'].indptr)
        assert_equal(P_I.indices, params['P_I'].indices)
        assert_equal(P_I.data, params['P_I'].data)

        A = A.tobsr(blocksize=(2, 2))
        Cpts = np.array([0])
        AggOp = csr_matrix(np.array([[1.]]))
        T = bsr_matrix(np.array([[1., 1.], [1., 2.]]), blocksize=(2, 2))
        params = get_Cpt_params(A, Cpts, AggOp, T)
        I_C = bsr_matrix(np.array([[1., 0.], [0., 1.]]), blocksize=(2, 2))
        I_F = bsr_matrix(np.array([[0., 0.], [0., 0.]]), blocksize=(2, 2))
        P_I = bsr_matrix(np.array([[1., 0.], [0., 1.]]), blocksize=(2, 2))
        assert_equal(np.array([0, 1]), params['Cpts'])
        assert_equal(np.array([]), params['Fpts'])
        assert_equal(I_C.indptr, params['I_C'].indptr)
        assert_equal(I_C.indices, params['I_C'].indices)
        assert_equal(I_C.data, params['I_C'].data)
        assert_equal(I_F.indptr, params['I_F'].indptr)
        assert_equal(I_F.indices, params['I_F'].indices)
        assert_equal(I_F.data, params['I_F'].data)
        assert_equal(P_I.indptr, params['P_I'].indptr)
        assert_equal(P_I.indices, params['P_I'].indices)
        assert_equal(P_I.data, params['P_I'].data)

        Cpts = np.array([])
        AggOp = csr_matrix(np.array([[1.]]))
        T = bsr_matrix(np.array([[1., 1.], [1., 2.]]), blocksize=(2, 2))
        params = get_Cpt_params(A, Cpts, AggOp, T)
        I_C = bsr_matrix(np.array([[0., 0.], [0., 0.]]), blocksize=(2, 2))
        I_F = bsr_matrix(np.array([[1., 0.], [0., 1.]]), blocksize=(2, 2))
        P_I = bsr_matrix(np.array([[0., 0.], [0., 0.]]), blocksize=(2, 2))
        assert_equal(np.array([]), params['Cpts'])
        assert_equal(np.array([0, 1]), params['Fpts'])
        assert_equal(I_C.indptr, params['I_C'].indptr)
        assert_equal(I_C.indices, params['I_C'].indices)
        assert_equal(I_C.data, params['I_C'].data)
        assert_equal(I_F.indptr, params['I_F'].indptr)
        assert_equal(I_F.indices, params['I_F'].indices)
        assert_equal(I_F.data, params['I_F'].data)
        assert_equal(P_I.indptr, params['P_I'].indptr)
        assert_equal(P_I.indices, params['P_I'].indices)
        assert_equal(P_I.data, params['P_I'].data)

        # Begin more "realistic" tests
        A = poisson((10,), format='csr')
        Cpts = np.array([3, 7])
        AggOp = ([[1., 0.],
                  [1., 0.],
                  [1., 0.],
                  [1., 0.],
                  [1., 0.],
                  [0., 1.],
                  [0., 1.],
                  [0., 1.],
                  [0., 1.],
                  [0., 1.]])
        AggOp = csr_matrix(AggOp)
        T = AggOp.copy().tobsr()

        # CSR Test
        params = get_Cpt_params(A, Cpts, AggOp, T)
        I_C = bsr_matrix((np.array([[[1.]], [[1.]]]),
                          np.array([3, 7]),
                          np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2])),
                         shape=(10, 10))
        I_F = bsr_matrix((np.array([[[1.]], [[1.]], [[1.]], [[1.]],
                                    [[1.]], [[1.]], [[1.]], [[1.]]]),
                          np.array([0, 1, 2, 4, 5, 6, 8, 9]),
                          np.array([0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8])),
                         shape=(10, 10))
        P_I = np.array([[0., 0.],
                        [0., 0.],
                        [0., 0.],
                        [1., 0.],
                        [0., 0.],
                        [0., 0.],
                        [0., 0.],
                        [0., 1.],
                        [0., 0.],
                        [0., 0.]])
        P_I = bsr_matrix(P_I, blocksize=(1, 1))
        Fpts = np.array([0, 1, 2, 4, 5, 6, 8, 9])
        assert_equal(Cpts, params['Cpts'])
        assert_equal(Fpts, params['Fpts'])
        assert_equal(I_C.indptr, params['I_C'].indptr)
        assert_equal(I_C.indices, params['I_C'].indices)
        assert_equal(I_C.data, params['I_C'].data)
        assert_equal(I_F.indptr, params['I_F'].indptr)
        assert_equal(I_F.indices, params['I_F'].indices)
        assert_equal(I_F.data, params['I_F'].data)
        assert_equal(P_I.indptr, params['P_I'].indptr)
        assert_equal(P_I.indices, params['P_I'].indices)
        assert_equal(P_I.data, params['P_I'].data)

        # BSR Test
        A = A.tobsr(blocksize=(2, 2))
        Cpts = np.array([1, 3])
        AggOp = ([[1., 0.],
                  [1., 0.],
                  [1., 0.],
                  [0., 1.],
                  [0., 1.]])
        AggOp = csr_matrix(AggOp)
        T = np.hstack((T.toarray(), T.toarray()))[:, [0, 2, 1, 3]]
        T = bsr_matrix(T, blocksize=(2, 2))
        params = get_Cpt_params(A, Cpts, AggOp, T)
        I_C = bsr_matrix((np.array([[[1., 0.], [0., 1.]],
                                    [[1., 0.], [0., 1.]]]),
                          np.array([1, 3]),
                          np.array([0, 0, 1, 1, 2, 2])),
                         shape=(10, 10))
        I_F = bsr_matrix((np.array([[[1., 0.], [0., 1.]],
                                    [[1., 0.], [0., 1.]],
                                    [[1., 0.], [0., 1.]]]),
                          np.array([0, 2, 4]),
                          np.array([0, 1, 1, 2, 2, 3])),
                         shape=(10, 10))
        P_I = np.array([[0., 0., 0., 0.],
                        [0., 0., 0., 0.],
                        [1., 0., 0., 0.],
                        [0., 1., 0., 0.],
                        [0., 0., 0., 0.],
                        [0., 0., 0., 0.],
                        [0., 0., 1., 0.],
                        [0., 0., 0., 1.],
                        [0., 0., 0., 0.],
                        [0., 0., 0., 0.]])
        P_I = bsr_matrix(P_I, blocksize=(2, 2))
        Fpts = np.array([0, 1, 4, 5, 8, 9])
        Cpts = np.array([2, 3, 6, 7])
        assert_equal(Cpts, params['Cpts'])
        assert_equal(Fpts, params['Fpts'])
        assert_equal(I_C.indptr, params['I_C'].indptr)
        assert_equal(I_C.indices, params['I_C'].indices)
        assert_equal(I_C.data, params['I_C'].data)
        assert_equal(I_F.indptr, params['I_F'].indptr)
        assert_equal(I_F.indices, params['I_F'].indices)
        assert_equal(I_F.data, params['I_F'].data)
        assert_equal(P_I.indptr, params['P_I'].indptr)
        assert_equal(P_I.indices, params['P_I'].indices)
        assert_equal(P_I.data, params['P_I'].data)

    def test_compute_BtBinv(self):
        # Trivially sized tests
        # 1x1x1
        T = np.array([[1.]])
        T = bsr_matrix(T, blocksize=(1, 1))
        B = np.array([[1.]])
        BtBinv = compute_BtBinv(B, T)
        answer = np.array([[[1.]]])
        assert_array_almost_equal(BtBinv, answer)

        T = np.array([[1.]])
        T = bsr_matrix(T, blocksize=(1, 1))
        B = np.array([[0.]])
        BtBinv = compute_BtBinv(B, T)
        answer = np.array([[[0.]]])
        assert_array_almost_equal(BtBinv, answer)

        T = np.array([[1.]])
        T = bsr_matrix(T, blocksize=(1, 1))
        B = np.array([[0.5]])
        BtBinv = compute_BtBinv(B, T)
        answer = np.array([[[4.]]])
        assert_array_almost_equal(BtBinv, answer)

        # 2x1x1
        T = np.array([[1., 0.], [1., 1.]])
        T = bsr_matrix(T, blocksize=(1, 1))
        B = np.array([[1.], [1.]])
        BtBinv = compute_BtBinv(B, T)
        answer = np.array([[[1.]], [[0.5]]])
        assert_array_almost_equal(BtBinv, answer)

        T = np.array([[1., 0.], [1., 1.]])
        T = bsr_matrix(T, blocksize=(1, 1))
        B = np.array([[0.], [1.]])
        BtBinv = compute_BtBinv(B, T)
        answer = np.array([[[0.]], [[1.]]])
        assert_array_almost_equal(BtBinv, answer)

        T = np.array([[1., 0.], [1., 1.]])
        T = bsr_matrix(T, blocksize=(2, 2))
        B = np.array([[0.], [2.]])
        BtBinv = compute_BtBinv(B, T)
        answer = np.array([[[0.25]]])
        assert_array_almost_equal(BtBinv, answer)

        T = np.array([[1., 0.], [1., 0.],
                      [0., .5], [0., .25]])
        T = bsr_matrix(T, blocksize=(1, 1))
        B = np.array([[1.], [2.]])
        BtBinv = compute_BtBinv(B, T)
        answer = np.array([[[1.]], [[1.]],
                           [[0.25]], [[0.25]]])
        assert_array_almost_equal(BtBinv, answer)

        T = np.array([[1., 0.], [0., .25]])
        T = bsr_matrix(T, blocksize=(1, 1))
        B = np.array([[1., 1.], [2., 1.]])
        BtBinv = compute_BtBinv(B, T)
        answer = np.array([[[0.25, 0.25], [0.25, 0.25]],
                           [[0.16, 0.08], [0.08, 0.04]]])
        assert_array_almost_equal(BtBinv, answer)

        T = np.array([[1., 0.], [0., .25]])
        T = bsr_matrix(T, blocksize=(2, 2))
        B = np.array([[1., 1.], [1., 1.]])
        BtBinv = compute_BtBinv(B, T)
        answer = np.array([[[0.125, 0.125],
                            [0.125, 0.125]]])
        assert_array_almost_equal(BtBinv, answer)

        # Simple BSR test
        T = np.array([[1., 1., 0., 0.],
                      [1., 1., 0., 0.],
                      [0., 0., 0.5, 0.5],
                      [0., 0., 0.25, 0.25]])
        T = bsr_matrix(T, blocksize=(2, 2))
        B = np.array([[1., 1.], [1., 2.], [1., 1.], [1., 3.]])
        BtBinv = compute_BtBinv(B, T)
        answer = np.array([[[5., -3.], [-3., 2.]],
                           [[2.5, -1.], [-1., 0.5]]])
        assert_array_almost_equal(BtBinv, answer)

    def test_eliminate_diag_dom_nodes(self):
        # Simple CSR test
        from pyamg.gallery import poisson
        A = poisson((4,), format='csr')
        C = eliminate_diag_dom_nodes(A, A.copy(), 1.1)
        answer = np.array([[1., 0., 0., 0.],
                           [0., 2., -1., 0.],
                           [0., -1., 2., 0.],
                           [0., 0., 0., 1.]])
        assert_array_almost_equal(C.toarray(), answer)

        # Simple BSR test
        from pyamg.gallery import poisson
        A = poisson((6,), format='csr')
        A.data[3] = 5.0
        A = A.tobsr((2, 2))
        C = poisson((3,), format='csr')
        C = eliminate_diag_dom_nodes(A, C, 1.1)
        answer = np.array([[1., 0., 0.],
                           [0., 2., -1.],
                           [0., -1., 2.]])
        assert_array_almost_equal(C.toarray(), answer)

    def test_remove_diagonal(self):
        from pyamg.gallery import poisson
        from pyamg.util.utils import remove_diagonal
        A = poisson((4,), format='csr')
        C = remove_diagonal(A)
        exact = np.array([[0., -1., 0., 0.],
                          [-1., 0., -1., 0.],
                          [0., -1., 0., -1.],
                          [0., 0., -1., 0.]])
        assert_array_almost_equal(C.toarray(), exact)

    def test_scale_rows_by_largest_entry(self):
        from pyamg.gallery import poisson
        from pyamg.util.utils import scale_rows_by_largest_entry
        A = poisson((4,), format='csr')
        A.data = np.array(A.data, dtype=complex)
        A.data[1] = 8.
        A = scale_rows_by_largest_entry(A)
        exact = np.array([[0.25, 1., 0., 0.],
                          [-0.5, 1., -0.5, 0.],
                          [0., -0.5, 1., -0.5],
                          [0., 0., -0.5, 1.]])
        assert_array_almost_equal(A.toarray(), exact)

    def test_filter_matrix_rows(self):
        from pyamg.util.utils import filter_matrix_rows
        A = csr_matrix(np.array([[0.24, -0.5, 0., 0.],
                                 [1., 1., 0.49, 0.],
                                 [0., -0.5, 1., -0.5]]))

        A2 = filter_matrix_rows(A, 0.5)
        exact = np.array([[0.0, -0.5, 0., 0.],
                          [1., 1., 0., 0.],
                          [0., -0.5, 1., -0.5]])
        assert_array_almost_equal(A2.toarray(), exact)

        A2 = A.copy()
        filter_matrix_rows(A2, 0.5, diagonal=True)
        exact = np.array([[0.24, -0.5, 0., 0.],
                          [1., 1., 0.0, 0.],
                          [0., -0.5, 1., -0.5]])
        assert_array_almost_equal(A2.toarray(), exact)

        A2 = A.copy()
        filter_matrix_rows(A2, 0.5, diagonal=True, lump=True)
        exact = np.array([[0.24, -0.5, 0., 0.],
                          [1., 1.49, 0.0, 0.],
                          [0., -0.5, 1., -0.5]])
        assert_array_almost_equal(A2.toarray(), exact)

    def test_filter_matrix_columns(self):
        from pyamg.util.utils import filter_matrix_columns
        A = csr_matrix(np.array([[0.24, 1., 0.],
                                 [-0.5, 1., -0.5],
                                 [0., 0.49, 1.],
                                 [0., 0., -0.5]]))
        A = filter_matrix_columns(A, 0.5)
        exact = np.array([[0., 1., 0.],
                          [-0.5, 1., -0.5],
                          [0., 0., 1.],
                          [0., 0., -0.5]])
        assert_array_almost_equal(A.toarray(), exact)

    def test_truncate_rows(self):
        from pyamg.util.utils import truncate_rows
        A = csr_matrix(np.array([[-0.24, -0.5, 0., 0.],
                                 [1., -1.1, 0.49, 0.1],
                                 [0., 0.4, 1., 0.5]]))
        Acopy = A.copy()
        Acopy = truncate_rows(Acopy, 4)
        assert_array_almost_equal(A.toarray(), Acopy.toarray())

        Acopy = A.copy()
        Acopy = truncate_rows(Acopy, 3)
        exact = np.array([[-0.24, -0.5, 0., 0.],
                          [1., -1.1, 0.49, 0.],
                          [0., 0.4, 1., 0.5]])
        assert_array_almost_equal(Acopy.toarray(), exact)

        Acopy = A.copy()
        Acopy = truncate_rows(Acopy, 2)
        exact = np.array([[-0.24, -0.5, 0., 0.],
                          [1., -1.1, 0., 0.],
                          [0., 0., 1., 0.5]])
        assert_array_almost_equal(Acopy.toarray(), exact)

        Acopy = A.copy()
        Acopy = truncate_rows(Acopy, 1)
        exact = np.array([[0., -0.5, 0., 0.],
                          [0., -1.1, 0., 0.],
                          [0., 0., 1., 0.]])
        assert_array_almost_equal(Acopy.toarray(), exact)

        Acopy = A.copy()
        Acopy = truncate_rows(Acopy, 0)
        exact = np.array([[0., 0., 0., 0.],
                          [0., 0., 0., 0.],
                          [0., 0., 0., 0.]])
        assert_array_almost_equal(Acopy.toarray(), exact)

    def test_scale_block_inverse(self):
        from pyamg.gallery import poisson
        from pyamg.util.utils import scale_block_inverse
        A = poisson((4,), format='csr')
        A, Dinv = scale_block_inverse(A, 2)
        A_stored = np.array([[1.,  0.,     -1./3.,  0.],
                             [0.,  1.,     -2./3.,  0.],
                             [0., -2./3.,  1.,      0.],
                             [0., -1./3.,  0.,      1.]])
        Dinv_stored = np.array([[2./3.,  1./3.,  0.,     0.],
                                [1./3.,  2./3.,  0.,     0.],
                                [0.,     0.,     2./3.,  1./3.],
                                [0.,     0.,     1./3.,  2./3.]])
        assert_array_almost_equal(A.toarray(), A_stored)
        assert_array_almost_equal(Dinv.toarray(), Dinv_stored)


class TestComplexUtils(TestCase):
    def test_diag_sparse(self):
        # check sparse -> array
        A = np.array([[-4-4.0j]])
        assert_equal(diag_sparse(csr_matrix(A)), [-4-4.0j])

        A = np.array([[1, 0, -5], [-2, 5-2.0j, 0]])
        assert_equal(diag_sparse(csr_matrix(A)), [1, 5-2.0j])

        # check array -> sparse
        A = np.array([[-4+1.0j]])
        assert_equal(diag_sparse(np.array([-4+1.0j])).toarray(),
                     csr_matrix(A).toarray())

        A = np.array([[1, 0], [0, 5-2.0j]])
        assert_equal(diag_sparse(np.array([1, 5-2.0j])).toarray(),
                     csr_matrix(A).toarray())

    def test_symmetric_rescaling(self):
        cases = []
        A = np.array([[5.5+1.0j, 3.5, 4.8],
                      [2., 9.9, 0.5-2.0j],
                      [6.5, 2.6, 5.7+1.0j]])
        A = csr_matrix(A)
        cases.append(A)
        P = diag_sparse([1, 0, 1.0j])
        cases.append(P*A*P)
        P = diag_sparse([0, 1+1.0j, 0])
        cases.append(P*A*P)

        for A in cases:
            D_sqrt, D_sqrt_inv, DAD = symmetric_rescaling(A)
            assert_almost_equal(diag_sparse(A) != 0, np.real(diag_sparse(DAD)))
            assert_almost_equal(diag_sparse(DAD), D_sqrt*D_sqrt_inv)

            D_sqrt, D_sqrt_inv = diag_sparse(D_sqrt), diag_sparse(D_sqrt_inv)
            assert_almost_equal((D_sqrt_inv*A*D_sqrt_inv).toarray(),
                                DAD.toarray())

    def test_symmetric_rescaling_sa(self):
        cases = []
        # case 1
        e = 1.0j*np.ones((5, 1)).ravel()
        data = [-1*e, 2*e, -1*e]
        A = 1.0j*spdiags(data, [-1, 0, 1], 5, 5).tocsr()
        B = e.copy().reshape(-1, 1)
        DAD_answer = np.array([[1., -0.5, 0., 0., 0.],
                               [-0.5, 1., -0.5, 0., 0.],
                               [0., -0.5, 1., -0.5, 0.],
                               [0., 0., -0.5, 1., -0.5],
                               [0., 0., 0., -0.5, 1.]])
        DB_answer = np.sqrt(2)*1.0j*e.reshape(-1, 1)
        #           matrix   B    BH   expected matrix  expected B  expected BH
        cases.append((A, B, None, DAD_answer, DB_answer, None))

        for case in cases:
            [A, B, BH, DAD_answer, DB_answer, DBH_answer] = case

            [DAD, DB, DBH] = symmetric_rescaling_sa(A, B, BH=BH)
            assert_array_almost_equal(DAD.toarray(), DAD_answer)
            assert_array_almost_equal(DB, DB_answer)
            if DBH_answer is not None:
                assert_array_almost_equal(DBH, DBH_answer)

    def test_get_diagonal(self):
        cases = []
        for i in range(1, 6):
            A = np.random.rand(i, i)
            Ai = A + 1.0j*np.random.rand(i, i)
            cases.append(csr_matrix(A))
            cases.append(csr_matrix(Ai))

        for A in cases:
            D_A = get_diagonal(A, norm_eq=False, inv=False)
            D_A_inv = get_diagonal(A, norm_eq=False, inv=True)
            D_AA = get_diagonal(A, norm_eq=1, inv=False)
            D_AA_inv = get_diagonal(A, norm_eq=1, inv=True)
            D_AA2 = get_diagonal(A, norm_eq=2, inv=False)
            D_AA_inv2 = get_diagonal(A, norm_eq=2, inv=True)

            D = np.diag(A.toarray())
            assert_almost_equal(D, D_A)
            D = 1.0/D
            assert_almost_equal(D, D_A_inv)

            D = np.diag((A.H*A).toarray())
            assert_almost_equal(D, D_AA)
            D = 1.0/D
            assert_almost_equal(D, D_AA_inv)

            D = np.diag((A*A.H).toarray())
            assert_almost_equal(D, D_AA2)
            D = 1.0/D
            assert_almost_equal(D, D_AA_inv2)

    def test_profile_solver(self):
        from scipy.sparse.linalg import cg
        from pyamg.gallery import poisson
        from pyamg.aggregation import smoothed_aggregation_solver

        A = poisson((100, 100), format='csr')
        A.data = A.data + 1e-5*np.random.rand(A.nnz)
        ml = smoothed_aggregation_solver(A)

        opts = []
        opts.append({})
        opts.append({'accel': cg, 'atol': 'legacy'})
        opts.append({'accel': cg, 'tol': 1e-10, 'atol': 'legacy'})

        for kwargs in opts:
            residuals = profile_solver(ml, **kwargs)
        del residuals

    def test_to_type(self):
        w = 1.2
        x = np.ones((5, 1))
        y = np.random.rand(3, 2)
        z = csr_matrix(np.random.rand(2, 2))
        inlist = [w, x, y, z]

        out = to_type(complex, inlist)
        for i in range(len(out)):
            assert out[i].dtype == complex
            if isspmatrix(out[i]):
                diff = np.ravel(out[i].data - inlist[i].data)
            else:
                diff = out[i] - inlist[i]
            assert_equal(max(np.abs(np.ravel(diff))), 0.0)

    def test_type_prep(self):
        w = 1.2
        x = np.ones((5, 1))
        y = np.random.rand(3, 2)
        z = csr_matrix(np.random.rand(2, 2))
        inlist = [w, x, y, z]

        out = type_prep(complex, inlist)
        for i in range(len(out)):
            assert out[i].dtype == complex
            assert not np.isscalar(out[i])
            if isspmatrix(out[i]):
                diff = np.ravel(out[i].data - inlist[i].data)
            else:
                diff = out[i] - inlist[i]
            assert_equal(max(np.abs(np.ravel(diff))), 0.0)

    def test_filter_operator(self):
        # Basic tests, with easy to compute answers
        # test one, the constant
        A = np.array([[1.+0.j, 1, 1], [1, 1, 1], [0, 1, 0],
                      [0, 1, 0], [0, 0, 1], [0, 0, 1]])
        C = np.array([[1.+0.j, 1, 0], [1, 1, 0], [0, 1, 0],
                      [0, 1, 0], [0, 0, 1], [0, 0, 1]])
        B = np.ones((3, 1)) + 0.j
        Bf = np.ones((6, 1)) + 1.0j * np.ones((6, 1))
        A_filter = filter_operator(csr_matrix(A), csr_matrix(C), B, Bf)
        A_filter = A_filter.toarray()
        A_known = np.array([[0.5+0.5j, 0.5+0.5j, 0.0+0.j],
                            [0.5+0.5j, 0.5+0.5j, 0.0+0.j],
                            [0.0+0.j, 1.0+1.j, 0.0+0.j],
                            [0.0+0.j, 1.0+1.j, 0.0+0.j],
                            [0.0+0.j, 0.0+0.j, 1.0+1.j],
                            [0.0+0.j, 0.0+0.j, 1.0+1.j]])
        assert_array_almost_equal(A_known, A_filter)

        # test two, the constant and linears
        # Note that for the rows with only one nonzero, Bf can't be
        # approximated exactly
        B = np.hstack((B, np.arange(B.shape[0]).reshape(-1, 1)))
        ww = np.arange(Bf.shape[0]).reshape(-1, 1) +\
            1.0j*np.arange(Bf.shape[0]).reshape(-1, 1)
        Bf = np.hstack((Bf, ww))
        A_filter = filter_operator(csr_matrix(A), csr_matrix(C), B, Bf)
        A_filter = A_filter.toarray()
        A_known = np.array([[1.0+1.j, 0.0+0.j, 0.0+0.j],
                            [0.0+0.j, 1.0+1.j, 0.0+0.j],
                            [0.0+0.j, 1.5+1.5j, 0.0+0.j],
                            [0.0+0.j, 2.0+2.j, 0.0+0.j],
                            [0.0+0.j, 0.0+0.j, 1.8+1.8j],
                            [0.0+0.j, 0.0+0.j, 2.2+2.2j]])
        assert_array_almost_equal(A_known, A_filter)

    def test_scale_T(self):
        # Test for one CSR and one BSR example
        T = np.array([[1.0, 0., 0.],
                      [0.5j, 0., 0.],
                      [0., 1., 0.],
                      [0., .5j, 0.],
                      [0., 0., 1.j],
                      [0., 0., 0.25]])
        T_answer = np.array([[0.-2.j, 0.+0.j, 0.+0.j],
                             [1.+0.j, 0.+0.j, 0.+0.j],
                             [0.+0.j, 1.+0.j, 0.+0.j],
                             [0.+0.j, 0.+0.5j, 0.+0.j],
                             [0.+0.j, 0.+0.j, 0.+4.j],
                             [0.+0.j, 0.+0.j, 1.+0.j]])
        P_I = np.array([[0., 0., 0.],
                        [1., 0., 0.],
                        [0., 1., 0.],
                        [0., 0., 0.],
                        [0., 0., 0.],
                        [0., 0., 1.]])
        I_F = np.array([[1., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 1., 0., 0.],
                        [0., 0., 0., 0., 1., 0.],
                        [0., 0., 0., 0., 0., 0.]])
        T_scaled = scale_T(bsr_matrix(T), bsr_matrix(P_I), bsr_matrix(I_F))
        T_scaled = T_scaled.toarray()
        assert_array_almost_equal(T_answer, T_scaled)

        # BSR test
        T = np.array([[1.j, 1., 0., 0.],
                      [0.5, 1., 0., 0.],
                      [1., 0., 0., 0.],
                      [0., 1., 0., 0.],
                      [0., 0., 2.j, 0.],
                      [0., 0., 0., 1.],
                      [0., 0., 1., 1.],
                      [0., 0., 1., 1.]])
        P_I = np.array([[0., 0., 0., 0.],
                        [0., 0., 0., 0.],
                        [1., 0., 0., 0.],
                        [0., 1., 0., 0.],
                        [0., 0., 1., 0.],
                        [0., 0., 0., 1.],
                        [0., 0., 0., 0.],
                        [0., 0., 0., 0.]])
        I_F = np.array([[1., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 1., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 0.],
                        [0., 0., 0., 0., 0., 0., 1., 0.],
                        [0., 0., 0., 0., 0., 0., 0., 1.]])
        T_answer = np.array([[0.0+1.j, 1.0+0.j, 0.0+0.j, 0.0+0.j],
                             [0.5+0.j, 1.0+0.j, 0.0+0.j, 0.0+0.j],
                             [1.0+0.j, 0.0+0.j, 0.0+0.j, 0.0+0.j],
                             [0.0+0.j, 1.0+0.j, 0.0+0.j, 0.0+0.j],
                             [0.0+0.j, 0.0+0.j, 1.0+0.j, 0.0+0.j],
                             [0.0+0.j, 0.0+0.j, 0.0+0.j, 1.0+0.j],
                             [0.0+0.j, 0.0+0.j, 0.0-0.5j, 1.0+0.j],
                             [0.0+0.j, 0.0+0.j, 0.0-0.5j, 1.0+0.j]])
        T = bsr_matrix(T, blocksize=(2, 2))
        P_I = bsr_matrix(P_I, blocksize=(2, 2))
        I_F = bsr_matrix(I_F, blocksize=(2, 2))
        T_scaled = scale_T(T, P_I, I_F).toarray()
        assert_array_almost_equal(T_answer, T_scaled)

    def test_compute_BtBinv(self):
        # Simple CSR test
        T = np.array([[1.j, 0.], [1., 0.],
                      [0., .5], [0., .25]])
        T = bsr_matrix(T, blocksize=(1, 1))
        B = np.array([[1.+1.j], [2.j]])
        BtBinv = compute_BtBinv(B, T)
        answer = np.array([[[0.50+0.j]], [[0.50+0.j]],
                           [[0.25+0.j]], [[0.25+0.j]]])
        assert_array_almost_equal(BtBinv, answer)

        # Simple BSR test
        T = np.array([[1., 0., 0., 1.],
                      [1., 0., 0., 1.],
                      [0., 0., 0.5, 0.],
                      [0., 0., 0.25, 0.]])
        T = bsr_matrix(T, blocksize=(2, 2))
        B = np.array([[1.j, 1.], [1.j, 3.], [1.j, 4.], [1.j, 2.]])
        BtBinv = compute_BtBinv(B, T)
        answer = np.array([[[1.5+0.j, 0.0+0.5j], [0.0-0.5j, 0.2+0.j]],
                           [[5.0+0.j, 0.0+1.5j], [0.0-1.5j, 0.5+0.j]]])
        assert_array_almost_equal(BtBinv, answer)

    def test_eliminate_diag_dom_nodes(self):
        # Simple CSR test
        from pyamg.gallery import poisson
        A = poisson((4,), format='csr')
        A.data = np.array(A.data, dtype=complex)
        A.data[0] = 1+5.0j
        C = eliminate_diag_dom_nodes(A, A.copy(), 1.1)
        answer = np.array([[1., 0., 0., 0.],
                           [0., 2., -1., 0.],
                           [0., -1., 2., 0.],
                           [0., 0., 0., 1.]])
        assert_array_almost_equal(C.toarray(), answer)

        # Simple BSR test
        from pyamg.gallery import poisson
        A = poisson((6,), format='csr')
        A.data = np.array(A.data, dtype=complex)
        A.data[3] = 1.0 + 5.0j
        A = A.tobsr((2, 2))
        C = poisson((3,), format='csr')
        C = eliminate_diag_dom_nodes(A, C, 1.1)
        answer = np.array([[1., 0., 0.],
                           [0., 2., -1.],
                           [0., -1., 2.]])
        assert_array_almost_equal(C.toarray(), answer)

    def test_remove_diagonal(self):
        from pyamg.gallery import poisson
        from pyamg.util.utils import remove_diagonal
        A = poisson((4,), format='csr')
        C = remove_diagonal(1.0j*A)
        exact = np.array([[0., -1., 0., 0.],
                          [-1., 0., -1., 0.],
                          [0., -1., 0., -1.],
                          [0., 0., -1., 0.]])
        assert_array_almost_equal(C.toarray(), 1.0j*exact)

    def test_scale_rows_by_largest_entry(self):
        from pyamg.gallery import poisson
        from pyamg.util.utils import scale_rows_by_largest_entry
        A = poisson((4,), format='csr')
        A.data = np.array(A.data, dtype=complex)
        A.data[1] = 3. + 2.j
        A = scale_rows_by_largest_entry(A)
        exact = np.array([[0.55470020+0.j, 0.83205029+0.5547002j, 0.0, 0.0],
                          [-0.50000000+0.j, 1.00000000+0.j, -0.5+0.j, 0.0+0.j],
                          [0.00000000+0.j, -0.50000000+0.j, 1.0+0.j, -0.5+0.j],
                          [0.00000000+0.j, 0.00000000+0.j, -0.5+0.j, 1.0+0.j]])
        assert_array_almost_equal(A.toarray(), exact)

    def test_scale_block_inverse(self):
        from pyamg.gallery import poisson
        from pyamg.util.utils import scale_block_inverse
        A = poisson((4,), format='csr', dtype=complex)
        A.data = 1.0j*A.data
        A, Dinv = scale_block_inverse(A, 2)
        A_stored = np.array([[1.,  0.,     -1./3.,  0.],
                             [0.,  1.,     -2./3.,  0.],
                             [0., -2./3.,  1.,      0.],
                             [0., -1./3.,  0.,      1.]])
        Dinv_stored = -1.0j*np.array([[2./3.,  1./3.,  0.,     0.],
                                      [1./3.,  2./3.,  0.,     0.],
                                      [0.,     0.,     2./3.,  1./3.],
                                      [0.,     0.,     1./3.,  2./3.]],
                                     dtype=complex)
        assert_array_almost_equal(A.toarray(), A_stored)
        assert_array_almost_equal(Dinv.toarray(), Dinv_stored)

    # note: no explicitly complex tests necessary for get_Cpt_params
    # def test_get_Cpt_params(self):


class TestLevelize(TestCase):

    def test_levelize_smooth_or_improve_candidates(self):
        from pyamg.util.utils import levelize_smooth_or_improve_candidates
        # test 1
        result = levelize_smooth_or_improve_candidates([('jacobi', {})], 5)
        assert_equal(result, [('jacobi', {}) for i in range(5)])
        # test 2
        result = levelize_smooth_or_improve_candidates('jacobi', 5)
        assert_equal(result, ['jacobi' for i in range(5)])
        # test 3
        result = levelize_smooth_or_improve_candidates(('jacobi', {}), 5)
        assert_equal(result, [('jacobi', {}) for i in range(5)])
        # test 4
        result = levelize_smooth_or_improve_candidates([('jacobi', {}), None],
                                                       5)
        assert_equal(result, [('jacobi', {}), None, None, None, None])

    def test_levelize_strength_or_aggregation(self):
        from pyamg.util.utils import levelize_strength_or_aggregation
        from pyamg.gallery import poisson
        A = poisson((100,), format='csr')
        # test 1
        max_levels, max_coarse, result = \
            levelize_strength_or_aggregation([('symmetric', {})], 5, 5)
        assert_equal(result, [('symmetric', {}) for i in range(4)])
        assert_equal(max_levels, 5)
        assert_equal(max_coarse, 5)
        # test 2
        max_levels, max_coarse, result =\
            levelize_strength_or_aggregation('symmetric', 5, 5)
        assert_equal(result, ['symmetric' for i in range(4)])
        assert_equal(max_levels, 5)
        assert_equal(max_coarse, 5)
        # test 3
        max_levels, max_coarse, result =\
            levelize_strength_or_aggregation(('symmetric', {}), 5, 5)
        assert_equal(result, [('symmetric', {}) for i in range(4)])
        assert_equal(max_levels, 5)
        assert_equal(max_coarse, 5)
        # test 4
        max_levels, max_coarse, result =\
            levelize_strength_or_aggregation([('symmetric', {}), None], 5, 5)
        assert_equal(result, [('symmetric', {}), None, None, None])
        assert_equal(max_levels, 5)
        assert_equal(max_coarse, 5)
        # test 5
        max_levels, max_coarse, result =\
            levelize_strength_or_aggregation(('predefined', {'C': A}), 5, 5)
        assert_array_equal(result, [('predefined', {'C': A})])
        assert_equal(max_levels, 2)
        assert_equal(max_coarse, 0)
        # test 6
        max_levels, max_coarse, result =\
            levelize_strength_or_aggregation([('predefined', {'C': A}),
                                              ('predefined', {'C': A})], 5, 5)
        assert_array_equal(result, [('predefined', {'C': A}),
                                    ('predefined', {'C': A})])
        assert_equal(max_levels, 3)
        assert_equal(max_coarse, 0)
        # test 7
        max_levels, max_coarse, result = \
            levelize_strength_or_aggregation(None, 5, 5)
        assert_equal(result, [(None, {}) for i in range(4)])
        assert_equal(max_levels, 5)
        assert_equal(max_coarse, 5)



"""Basic PyAMG demo showing AMG standalone convergence versus preconditioned CG with AMG."""

import numpy as np

from .laplacian import poisson
from ..aggregation.aggregation import smoothed_aggregation_solver


def demo():
    """Outline basic demo."""
    A = poisson((100, 100), format='csr')  # 2D FD Poisson problem
    B = None                               # no near-null spaces guesses for SA
    b = np.random.rand(A.shape[0], 1)      # a random right-hand side

    # use AMG based on Smoothed Aggregation (SA) and display info
    mls = smoothed_aggregation_solver(A, B=B)
    print(mls)

    # Solve Ax=b with no acceleration ('standalone' solver)
    standalone_residuals = []
    x = mls.solve(b, tol=1e-10, accel=None, residuals=standalone_residuals)

    # Solve Ax=b with Conjugate Gradient (AMG as a preconditioner to CG)
    accelerated_residuals = []
    x = mls.solve(b, tol=1e-10, accel='cg', residuals=accelerated_residuals)
    del x

    # Compute relative residuals
    standalone_residuals = np.array(standalone_residuals) / standalone_residuals[0]
    accelerated_residuals = np.array(accelerated_residuals) / accelerated_residuals[0]

    # Compute (geometric) convergence factors
    factor1 = standalone_residuals[-1]**(1.0/len(standalone_residuals))
    factor2 = accelerated_residuals[-1]**(1.0/len(accelerated_residuals))

    print(f'                     MG convergence factor: {factor1}')
    print(f'MG with CG acceleration convergence factor: {factor2}')

    # Plot convergence history
    try:
        import matplotlib.pyplot as plt  # pylint: disable=import-outside-toplevel
        plt.figure()
        plt.title('Convergence History')
        plt.xlabel('Iteration')
        plt.ylabel('Relative Residual')
        plt.semilogy(standalone_residuals, label='Standalone', linestyle='-', marker='o')
        plt.semilogy(accelerated_residuals, label='Accelerated', linestyle='-', marker='s')
        plt.legend()
        plt.show()
    except ImportError:
        print('\nNote: matplotlib is needed for plotting.')



"""Test classical AMG."""
import warnings

import numpy as np

from numpy.testing import TestCase, assert_equal, assert_almost_equal, \
    assert_array_almost_equal

from scipy.sparse import csr_matrix, coo_matrix, SparseEfficiencyWarning

from pyamg.gallery import poisson, load_example
from pyamg.strength import classical_strength_of_connection

from pyamg.classical import split
from pyamg.classical.classical import ruge_stuben_solver
from pyamg.classical.interpolate import direct_interpolation, \
    classical_interpolation


class TestRugeStubenFunctions(TestCase):
    def setUp(self):
        self.cases = []

        # random matrices
        np.random.seed(0)
        for N in [2, 3, 5]:
            self.cases.append(csr_matrix(np.random.rand(N, N)))

        # Poisson problems in 1D and 2D
        for N in [2, 3, 5, 7, 10, 11, 19]:
            self.cases.append(poisson((N,), format='csr'))
        for N in [2, 3, 5, 7, 10, 11]:
            self.cases.append(poisson((N, N), format='csr'))

        for name in ['knot', 'airfoil', 'bar']:
            ex = load_example(name)
            self.cases.append(ex['A'].tocsr())

    def test_RS_splitting(self):
        for A in self.cases:
            S = classical_strength_of_connection(A, 0.0)

            splitting = split.RS(S)

            assert splitting.min() >= 0     # could be all 1s
            assert_equal(splitting.max(), 1)

            S.data[:] = 1

            # check that all F-nodes are strongly connected to a C-node
            assert (splitting + S*splitting).min() > 0

            # THIS IS NOT STRICTLY ENFORCED!
            # check that all strong connections S[i, j] satisfy either:
            # (0) i is a C-node
            # (1) j is a C-node
            # (2) k is a C-node and both i and j are strongly connected to k
            #
            # X = S.tocoo()

            # remove C->F edges (i.e. S[i, j] where (0) holds)
            # mask = splitting[X.row] == 0
            # X.row  = X.row[mask]
            # X.col  = X.col[mask]
            # X.data = X.data[mask]

            # remove F->C edges (i.e. S[i, j] where (1) holds)
            # mask = splitting[X.col] == 0
            # X.row  = X.row[mask]
            # X.col  = X.col[mask]
            # X.data = X.data[mask]

            # X now consists of strong F->F edges only
            #
            # (S * S.T)[i, j] is the # of C nodes on which both i and j
            # strongly depend (i.e. the number of k's where (2) holds)
            # Y = (S*S.T) - X
            # assert(Y.nnz == 0 or Y.data.min() > 0)

    def test_cljp_splitting(self):
        for A in self.cases:
            S = classical_strength_of_connection(A, 0.0)

            splitting = split.CLJP(S)

            assert splitting.min() >= 0     # could be all 1s
            assert_equal(splitting.max(), 1)

            S.data[:] = 1

            # check that all F-nodes are strongly connected to a C-node
            assert (splitting + S*splitting).min() > 0

    def test_cljpc_splitting(self):
        for A in self.cases:
            S = classical_strength_of_connection(A, 0.0)

            splitting = split.CLJPc(S)

            assert splitting.min() >= 0     # could be all 1s
            assert_equal(splitting.max(), 1)

            S.data[:] = 1

            # check that all F-nodes are strongly connected to a C-node
            assert (splitting + S*splitting).min() > 0

    def test_direct_interpolation(self):
        for A in self.cases:

            S = classical_strength_of_connection(A, 0.0)
            splitting = split.RS(S)

            result = direct_interpolation(A, S, splitting)
            expected = reference_direct_interpolation(A, S, splitting)

            assert_almost_equal(result.toarray(), expected.toarray())

    def test_classical_interpolation(self):
        for A in self.cases:
            # the reference code is very slow, so just take a small block of A
            mini = min(100, A.shape[0])
            A = ((A.tocsr()[0:mini, :])[:, 0:mini]).tocsr()

            S = classical_strength_of_connection(A, 0.0)
            splitting = split.RS(S, second_pass=True)

            result = classical_interpolation(A, S, splitting, modified=False)
            expected = reference_classical_interpolation(A, S, splitting)

            # elasticity produces large entries, so normalize
            Diff = result - expected
            Diff.data = abs(Diff.data)
            expected.data = 1./abs(expected.data)
            Diff = Diff.multiply(expected)
            Diff.data[Diff.data < 1e-7] = 0.0
            Diff.eliminate_zeros()
            assert (Diff.nnz == 0)

    def test_remove_strong_FF_connections(self):
        from pyamg import amg_core
        # test removing an F-F connection without any strong C in between (4--2),
        # while keeping the F-f connection (4--6) which does have a strong C in between
        C = csr_matrix(np.array([[2., 1., 0., 0., 0., 0.],
                                 [1., 2., 1., 1., 0., 0.],
                                 [0., 1., 2., 0., 0., 0.],
                                 [0., 1., 0., 2., 1., 1.],
                                 [0., 0., 0., 1., 2., 1.],
                                 [0., 0., 0., 1., 1., 2.]]))
        splitting = np.array([1, 0, 1, 0, 1, 0], dtype=C.indices.dtype)
        amg_core.remove_strong_FF_connections(6, C.indptr, C.indices,
                                              C.data, splitting)

        exact = np.array([[2., 1., 0., 0., 0., 0.],
                          [1., 2., 1., 0., 0., 0.],
                          [0., 1., 2., 0., 0., 0.],
                          [0., 0., 0., 2., 1., 1.],
                          [0., 0., 0., 1., 2., 1.],
                          [0., 0., 0., 1., 1., 2.]])
        assert_array_almost_equal(C.toarray(), exact)


class TestSolverPerformance(TestCase):
    def test_poisson(self):
        cases = []

        cases.append((500,))
        cases.append((250, 250))
        cases.append((25, 25, 25))

        for case in cases:
            A = poisson(case, format='csr')

            for interp in ['direct',
                           ('classical', {'modified': False}),
                           ('classical', {'modified': True})]:

                np.random.seed(0)  # make tests repeatable
                x = np.random.rand(A.shape[0])
                b = A*np.random.rand(A.shape[0])  # zeros_like(x)

                ml = ruge_stuben_solver(A, interpolation=interp, max_coarse=50)

                res = []
                x_sol = ml.solve(b, x0=x, maxiter=20, tol=1e-12,
                                 residuals=res)
                del x_sol

                avg_convergence_ratio = (res[-1]/res[0])**(1.0/len(res))
                assert (avg_convergence_ratio < 0.20)

    def test_matrix_formats(self):
        warnings.simplefilter('ignore', SparseEfficiencyWarning)

        # Do dense, csr, bsr and csc versions of A all yield the same solver
        A = poisson((7, 7), format='csr')
        cases = [A.tobsr(blocksize=(1, 1))]
        cases.append(A.tocsc())
        cases.append(A.toarray())

        rs_old = ruge_stuben_solver(A, max_coarse=10)
        for AA in cases:
            rs_new = ruge_stuben_solver(AA, max_coarse=10)
            Ac_old = rs_old.levels[-1].A.toarray()
            Ac_new = rs_new.levels[-1].A.toarray()
            assert np.abs(np.ravel(Ac_old - Ac_new)).max() < 0.01
            rs_old = rs_new


#   reference implementations for unittests  #
def reference_direct_interpolation(A, S, splitting):

    # Interpolation weights are computed based on entries in A, but subject to
    # the sparsity pattern of C.  So, we copy the entries of A into the
    # sparsity pattern of C.
    S = S.copy()
    S.data[:] = 1.0
    S = S.multiply(A)

    A = coo_matrix(A)
    S = coo_matrix(S)

    # remove diagonals
    mask = S.row != S.col
    S.row = S.row[mask]
    S.col = S.col[mask]
    S.data = S.data[mask]

    # strong C points
    c_mask = splitting[S.col] == 1
    C_s = coo_matrix((S.data[c_mask], (S.row[c_mask], S.col[c_mask])),
                     shape=S.shape)

    # strong F points
    # f_mask = ~c_mask
    # F_s = coo_matrix((S.data[f_mask], (S.row[f_mask], S.col[f_mask])),
    #                shape=S.shape)

    # split A in to + and -
    mask = (A.data > 0) & (A.row != A.col)
    A_pos = coo_matrix((A.data[mask], (A.row[mask], A.col[mask])),
                       shape=A.shape)
    mask = (A.data < 0) & (A.row != A.col)
    A_neg = coo_matrix((A.data[mask], (A.row[mask], A.col[mask])),
                       shape=A.shape)

    # split C_S in to + and -
    mask = C_s.data > 0
    C_s_pos = coo_matrix((C_s.data[mask], (C_s.row[mask], C_s.col[mask])),
                         shape=A.shape)
    mask = ~mask
    C_s_neg = coo_matrix((C_s.data[mask], (C_s.row[mask], C_s.col[mask])),
                         shape=A.shape)

    sum_strong_pos = np.ravel(C_s_pos.sum(axis=1))
    sum_strong_neg = np.ravel(C_s_neg.sum(axis=1))

    sum_all_pos = np.ravel(A_pos.sum(axis=1))
    sum_all_neg = np.ravel(A_neg.sum(axis=1))

    diag = A.diagonal()

    mask = (sum_strong_neg != 0.0)
    alpha = np.zeros_like(sum_all_neg)
    alpha[mask] = sum_all_neg[mask] / sum_strong_neg[mask]

    mask = (sum_strong_pos != 0.0)
    beta = np.zeros_like(sum_all_pos)
    beta[mask] = sum_all_pos[mask] / sum_strong_pos[mask]

    mask = sum_strong_pos == 0
    diag[mask] += sum_all_pos[mask]
    beta[mask] = 0

    C_s_neg.data *= -alpha[C_s_neg.row]/diag[C_s_neg.row]
    C_s_pos.data *= -beta[C_s_pos.row]/diag[C_s_pos.row]

    C_rows = splitting.nonzero()[0]
    C_inject = coo_matrix((np.ones(sum(splitting)), (C_rows, C_rows)),
                          shape=A.shape)

    P = C_s_neg.tocsr() + C_s_pos.tocsr() + C_inject.tocsr()

    splitting_map = np.concatenate(([0], np.cumsum(splitting)))
    P = csr_matrix((P.data, splitting_map[P.indices], P.indptr),
                   shape=(P.shape[0], splitting_map[-1]))

    return P


# strength has zero diagonal...?
def reference_classical_interpolation(A, S, splitting):

    # this routine only tests the computation of the "weights" the computation
    # of the sparsity pattern is the same as for direct interpolation, and is
    # tested through the reference_direct_interpolation routine.
    from pyamg import amg_core
    S = S.copy()
    S.data[:] = 1.0
    S = S.multiply(A)
    Pp = np.empty_like(A.indptr)
    amg_core.rs_direct_interpolation_pass1(A.shape[0], S.indptr, S.indices,
                                           splitting, Pp)
    nnz = Pp[-1]
    Pj = np.empty(nnz, dtype=Pp.dtype)
    Px = np.empty(nnz, dtype=A.dtype)
    SD = S.diagonal()
    F_NODE = 0
    C_NODE = 1

    # Now, we implement the second pass in Python to double check the C++ code
    for i in range(A.shape[0]):
        # If node is is a C-point, do injection
        if (splitting[i] == C_NODE):
            Pj[Pp[i]] = i
            Px[Pp[i]] = 1

        # Else compute classical interpolation weight
        else:
            rowstartA = A.indptr[i]
            rowendA = A.indptr[i+1]
            rowstartS = S.indptr[i]
            rowendS = S.indptr[i+1]

            # Denominator = a_ii + sum_{m in weak connections} a_im
            denominator = sum(A.data[rowstartA:rowendA])
            denominator -= sum(S.data[rowstartS:rowendS])
            denominator += SD[i]

            # Compute interpolation weights from strongly connected C-points
            nnz = Pp[i]
            for jj in range(rowstartS, rowendS):
                Sj = S.indices[jj]
                if (splitting[Sj] == C_NODE) and (Sj != i):
                    Pj[nnz] = Sj
                    numerator = S.data[jj]
                    for kk in range(rowstartS, rowendS):
                        Sk = S.indices[kk]
                        if (splitting[Sk] == F_NODE) and (Sk != i):
                            inner_denominator = 0.0
                            for ll in range(rowstartS, rowendS):
                                Sl = S.indices[ll]
                                if (splitting[Sl] == C_NODE) and (Sl != i):
                                    for search_ind in range(A.indptr[Sk], A.indptr[Sk+1]):
                                        if (A.indices[search_ind] == Sl):
                                            inner_denominator += A.data[search_ind]

                            for search_ind in range(A.indptr[Sk], A.indptr[Sk+1]):
                                if (A.indices[search_ind] == Sj) and \
                                   (inner_denominator != 0.0):
                                    numerator += \
                                        (S.data[kk]*A.data[search_ind]/inner_denominator)

                    Px[nnz] = -numerator/denominator
                    nnz += 1

    reorder = np.zeros((A.shape[0],))
    cumulative = 0
    for i in range(A.shape[0]):
        reorder[i] = cumulative
        cumulative += splitting[i]

    for i in range(Pp[A.shape[0]]):
        Pj[i] = reorder[Pj[i]]

    return csr_matrix((Px, Pj, Pp))



"""Test Pairwise AMG."""
import numpy as np

from numpy.testing import TestCase
from pyamg.aggregation import pairwise_solver
from pyamg.gallery import poisson, linear_elasticity, load_example


class TestPairwise(TestCase):

    def test_spd(self):
        cases = []
        cases.append(poisson((500,), format='csr'))
        cases.append(poisson((50, 50), format='csr'))
        cases.append(linear_elasticity((7, 7), format='bsr')[0])
        cases.append(load_example('airfoil')['A'].tocsr())

        for A in cases:
            for agg, expected in [(('pairwise', {'theta': 0.25, 'matchings': 2}), 0.75),
                                  (('pairwise', {'theta': 0.10, 'matchings': 2}), 0.75),
                                  (('pairwise', {'theta': 0.25, 'matchings': 1}), 0.75),
                                  (('pairwise', {'theta': 0.10, 'matchings': 1}), 0.75)]:

                np.random.seed(0)  # make tests repeatable
                x = np.random.rand(A.shape[0])
                b = A*np.random.rand(A.shape[0])

                ml = pairwise_solver(A, aggregate=agg, max_coarse=10)

                res = []
                x_sol = ml.solve(b, x0=x, maxiter=20, tol=1e-12,
                                 residuals=res)
                del x_sol

                avg_convergence_ratio = (res[-1]/res[0])**(1.0/len(res))
                assert (avg_convergence_ratio < expected)



"""Support for aggregation-based AMG."""


from warnings import warn
import numpy as np
from scipy.sparse import csr_matrix, isspmatrix_csr, isspmatrix_bsr,\
    SparseEfficiencyWarning

from ..multilevel import MultilevelSolver
from ..relaxation.smoothing import change_smoothers
from ..relaxation.utils import relaxation_as_linear_operator
from ..util.utils import scale_T, get_Cpt_params, \
    eliminate_diag_dom_nodes, get_blocksize, \
    levelize_strength_or_aggregation, \
    levelize_smooth_or_improve_candidates
from ..strength import classical_strength_of_connection,\
    symmetric_strength_of_connection, evolution_strength_of_connection,\
    energy_based_strength_of_connection, distance_strength_of_connection,\
    algebraic_distance, affinity_distance
from .aggregate import standard_aggregation, naive_aggregation, \
    lloyd_aggregation, pairwise_aggregation
from .tentative import fit_candidates
from .smooth import energy_prolongation_smoother


def rootnode_solver(A, B=None, BH=None,
                    symmetry='hermitian', strength='symmetric',
                    aggregate='standard', smooth='energy',
                    presmoother=('block_gauss_seidel',
                                 {'sweep': 'symmetric'}),
                    postsmoother=('block_gauss_seidel',
                                  {'sweep': 'symmetric'}),
                    improve_candidates=('block_gauss_seidel',
                                        {'sweep': 'symmetric',
                                         'iterations': 4}),
                    max_levels=10, max_coarse=10,
                    diagonal_dominance=False, keep=False, **kwargs):
    """Create a multilevel solver using root-node based Smoothed Aggregation (SA).

    See the notes below, for the major differences with the classical-style
    smoothed aggregation solver in aggregation.smoothed_aggregation_solver.

    Parameters
    ----------
    A : csr_matrix, bsr_matrix
        Sparse NxN matrix in CSR or BSR format

    B : None, array_like
        Right near-nullspace candidates stored in the columns of an NxK array.
        K must be >= the blocksize of A (see reference [2011OlScTu]_). The default value
        B=None is equivalent to choosing the constant over each block-variable,
        B=np.kron(np.ones((A.shape[0]/get_blocksize(A), 1)), np.eye(get_blocksize(A)))

    BH : None, array_like
        Left near-nullspace candidates stored in the columns of an NxK array.
        BH is only used if symmetry='nonsymmetric'.  K must be >= the
        blocksize of A (see reference [2011OlScTu]_). The default value B=None is
        equivalent to choosing the constant over each block-variable,
        B=np.kron(np.ones((A.shape[0]/get_blocksize(A), 1)), np.eye(get_blocksize(A)))

    symmetry : string
        'symmetric' refers to both real and complex symmetric
        'hermitian' refers to both complex Hermitian and real Hermitian
        'nonsymmetric' i.e. nonsymmetric in a hermitian sense
        Note that for the strictly real case, symmetric and hermitian are
        the same
        Note that this flag does not denote definiteness of the operator.

    strength : list
        Method used to determine the strength of connection between unknowns of
        the linear system.  Method-specific parameters may be passed in using a
        tuple, e.g. strength=('symmetric',{'theta' : 0.25 }). If strength=None,
        all nonzero entries of the matrix are considered strong.

    aggregate : list
        Method used to aggregate nodes.

    smooth : list
        Method used to smooth the tentative prolongator.  Method-specific
        parameters may be passed in using a tuple, e.g.  smooth=
        ('energy',{'krylov' : 'gmres'}).  Only 'energy' and None are valid
        prolongation smoothing options.

    presmoother : tuple, string, list
        Defines the presmoother for the multilevel cycling.  The default block
        Gauss-Seidel option defaults to point-wise Gauss-Seidel, if the matrix
        is CSR or is a BSR matrix with blocksize of 1.  See notes below for
        varying this parameter on a per level basis.

    postsmoother : tuple, string, list
        Same as presmoother, except defines the postsmoother.

    improve_candidates : tuple, string, list
        The ith entry defines the method used to improve the candidates B on
        level i.  If the list is shorter than max_levels, then the last entry
        will define the method for all levels lower.  If tuple or string, then
        this single relaxation descriptor defines improve_candidates on all
        levels.
        The list elements are relaxation descriptors of the form used for
        presmoother and postsmoother.  A value of None implies no action on B.

    max_levels : integer
        Maximum number of levels to be used in the multilevel solver.

    max_coarse : integer
        Maximum number of variables permitted on the coarse grid.

    diagonal_dominance : bool, tuple
        If True (or the first tuple entry is True), then avoid coarsening
        diagonally dominant rows.  The second tuple entry requires a
        dictionary, where the key value 'theta' is used to tune the diagonal
        dominance threshold.

    keep : bool
        Flag to indicate keeping extra operators in the hierarchy for
        diagnostics.  For example, if True, then strength of connection (C),
        tentative prolongation (T), aggregation (AggOp), and arrays
        storing the C-points (Cpts) and F-points (Fpts) are kept at
        each level.

    Other Parameters
    ----------------
    cycle_type : ['V','W','F']
        Structrure of multigrid cycle
    coarse_solver : ['splu', 'lu', 'cholesky, 'pinv', 'gauss_seidel', ... ]
        Solver used at the coarsest level of the MG hierarchy.
        Optionally, may be a tuple (fn, args), where fn is a string such as
        ['splu', 'lu', ...] or a callable function, and args is a dictionary of
        arguments to be passed to fn.

    Returns
    -------
    ml : MultilevelSolver
        Multigrid hierarchy of matrices and prolongation operators

    See Also
    --------
    MultilevelSolver, aggregation.smoothed_aggregation_solver,
    classical.ruge_stuben_solver

    Notes
    -----
         - Root-node style SA differs from classical SA primarily by preserving
           and identity block in the interpolation operator, P.  Each aggregate
           has a 'root-node' or 'center-node' associated with it, and this
           root-node is injected from the coarse grid to the fine grid.  The
           injection corresponds to the identity block.

         - Only smooth={'energy', None} is supported for prolongation
           smoothing.  See reference [2011OlScTu]_ below for more details on why the
           'energy' prolongation smoother is the natural counterpart to
           root-node style SA.

         - The additional parameters are passed through as arguments to
           MultilevelSolver.  Refer to pyamg.MultilevelSolver for additional
           documentation.

         - At each level, four steps are executed in order to define the coarser
           level operator.

           1. Matrix A is given and used to derive a strength matrix, C.

           2. Based on the strength matrix, indices are grouped or aggregated.

           3. The aggregates define coarse nodes and a tentative prolongation
              operator T is defined by injection

           4. The tentative prolongation operator is smoothed by a relaxation
              scheme to improve the quality and extent of interpolation from the
              aggregates to fine nodes.

         - The parameters smooth, strength, aggregate, presmoother, postsmoother
           can be varied on a per level basis.  For different methods on
           different levels, use a list as input so that the i-th entry defines
           the method at the i-th level.  If there are more levels in the
           hierarchy than list entries, the last entry will define the method
           for all levels lower.

           Examples are:
           smooth=[('jacobi', {'omega':1.0}), None, 'jacobi']
           presmoother=[('block_gauss_seidel', {'sweep':symmetric}), 'sor']
           aggregate=['standard', 'naive', 'lloyd', 'pairwise']
           strength=[('symmetric', {'theta':0.25}), ('symmetric', {'theta':0.08})]

         - Predefined strength of connection and aggregation schemes can be
           specified.  These options are best used together, but aggregation can
           be predefined while strength of connection is not.

           For predefined strength of connection, use a list consisting of
           tuples of the form ('predefined', {'C' : C0}), where C0 is a
           csr_matrix and each degree-of-freedom in C0 represents a supernode.
           For instance to predefine a three-level hierarchy, use
           [('predefined', {'C' : C0}), ('predefined', {'C' : C1}) ].

           Similarly for predefined aggregation, use a list of tuples.  For
           instance to predefine a three-level hierarchy, use [('predefined',
           {'AggOp' : Agg0}), ('predefined', {'AggOp' : Agg1}) ], where the
           dimensions of A, Agg0 and Agg1 are compatible, i.e.  Agg0.shape[1] ==
           A.shape[0] and Agg1.shape[1] == Agg0.shape[0].  Each AggOp is a
           csr_matrix.

           Because this is a root-nodes solver, if a member of the predefined
           aggregation list is predefined, it must be of the form
           ('predefined', {'AggOp' : Agg, 'Cnodes' : Cnodes}).

    Examples
    --------
    >>> from pyamg import rootnode_solver
    >>> from pyamg.gallery import poisson
    >>> from scipy.sparse.linalg import cg
    >>> import numpy as np
    >>> A = poisson((100, 100), format='csr')           # matrix
    >>> b = np.ones((A.shape[0]))                   # RHS
    >>> ml = rootnode_solver(A)                     # AMG solver
    >>> M = ml.aspreconditioner(cycle='V')             # preconditioner
    >>> x, info = cg(A, b, tol=1e-8, maxiter=30, M=M)   # solve with CG

    References
    ----------
    .. [1996VaMa] Vanek, P. and Mandel, J. and Brezina, M.,
       "Algebraic Multigrid by Smoothed Aggregation for
       Second and Fourth Order Elliptic Problems",
       Computing, vol. 56, no. 3, pp. 179--196, 1996.
       http://citeseer.ist.psu.edu/vanek96algebraic.html
    .. [2011OlScTu] Olson, L. and Schroder, J. and Tuminaro, R.,
       "A general interpolation strategy for algebraic
       multigrid using energy minimization", SIAM Journal
       on Scientific Computing (SISC), vol. 33, pp.
       966--991, 2011.

    """
    if not (isspmatrix_csr(A) or isspmatrix_bsr(A)):
        try:
            A = csr_matrix(A)
            warn('Implicit conversion of A to CSR',
                 SparseEfficiencyWarning)
        except BaseException as e:
            raise TypeError('Argument A must have type csr_matrix, '
                            'bsr_matrix, or be convertible to csr_matrix') from e

    A = A.asfptype()

    if symmetry not in ('symmetric', 'hermitian', 'nonsymmetric'):
        raise ValueError('Expected "symmetric", "nonsymmetric" '
                         'or "hermitian" for the symmetry parameter.')
    A.symmetry = symmetry

    if A.shape[0] != A.shape[1]:
        raise ValueError('expected square matrix')
    # Right near nullspace candidates use constant for each variable as default
    if B is None:
        B = np.kron(np.ones((int(A.shape[0]/get_blocksize(A)), 1), dtype=A.dtype),
                    np.eye(get_blocksize(A)))
    else:
        B = np.asarray(B, dtype=A.dtype)
        if len(B.shape) == 1:
            B = B.reshape(-1, 1)
        if B.shape[0] != A.shape[0]:
            raise ValueError('The near null-space modes B have incorrect \
                              dimensions for matrix A')
        if B.shape[1] < get_blocksize(A):
            raise ValueError('B.shape[1] must be >= the blocksize of A')

    # Left near nullspace candidates
    if A.symmetry == 'nonsymmetric':
        if BH is None:
            BH = B.copy()
        else:
            BH = np.asarray(BH, dtype=A.dtype)
            if len(BH.shape) == 1:
                BH = BH.reshape(-1, 1)
            if BH.shape[1] != B.shape[1]:
                raise ValueError('The number of left and right near \
                                  null-space modes B and BH, must be equal')
            if BH.shape[0] != A.shape[0]:
                raise ValueError('The near null-space modes BH have \
                                  incorrect dimensions for matrix A')

    # Levelize the user parameters, so that they become lists describing the
    # desired user option on each level.
    max_levels, max_coarse, strength =\
        levelize_strength_or_aggregation(strength, max_levels, max_coarse)
    max_levels, max_coarse, aggregate =\
        levelize_strength_or_aggregation(aggregate, max_levels, max_coarse)
    improve_candidates =\
        levelize_smooth_or_improve_candidates(improve_candidates, max_levels)
    smooth = levelize_smooth_or_improve_candidates(smooth, max_levels)

    # Construct multilevel structure
    levels = []
    levels.append(MultilevelSolver.Level())
    levels[-1].A = A          # matrix

    # Append near nullspace candidates
    levels[-1].B = B          # right candidates
    if A.symmetry == 'nonsymmetric':
        levels[-1].BH = BH    # left candidates

    while len(levels) < max_levels and \
            int(levels[-1].A.shape[0]/get_blocksize(levels[-1].A)) > max_coarse:
        _extend_hierarchy(levels, strength, aggregate, smooth,
                          improve_candidates, diagonal_dominance, keep)

    ml = MultilevelSolver(levels, **kwargs)
    change_smoothers(ml, presmoother, postsmoother)
    return ml


def _extend_hierarchy(levels, strength, aggregate, smooth, improve_candidates,
                      diagonal_dominance=False, keep=True):
    """Extend the multigrid hierarchy.

    Service routine to implement the strength of connection, aggregation,
    tentative prolongation construction, and prolongation smoothing.  Called by
    smoothed_aggregation_solver.

    """
    def unpack_arg(v):
        if isinstance(v, tuple):
            return v[0], v[1]
        return v, {}

    A = levels[-1].A
    B = levels[-1].B
    if A.symmetry == 'nonsymmetric':
        AH = A.H.asformat(A.format)
        BH = levels[-1].BH

    # Compute the strength-of-connection matrix C, where larger
    # C[i, j] denote stronger couplings between i and j.
    fn, kwargs = unpack_arg(strength[len(levels)-1])
    if fn == 'symmetric':
        C = symmetric_strength_of_connection(A, **kwargs)
    elif fn == 'classical':
        C = classical_strength_of_connection(A, **kwargs)
    elif fn == 'distance':
        C = distance_strength_of_connection(A, **kwargs)
    elif fn in ('ode', 'evolution'):
        if 'B' in kwargs:
            C = evolution_strength_of_connection(A, **kwargs)
        else:
            C = evolution_strength_of_connection(A, B, **kwargs)
    elif fn == 'energy_based':
        C = energy_based_strength_of_connection(A, **kwargs)
    elif fn == 'predefined':
        C = kwargs['C'].tocsr()
    elif fn == 'algebraic_distance':
        C = algebraic_distance(A, **kwargs)
    elif fn == 'affinity':
        C = affinity_distance(A, **kwargs)
    elif fn is None:
        C = A.tocsr()
    else:
        raise ValueError(f'Unrecognized strength of connection method: {str(fn)}')

    # Avoid coarsening diagonally dominant rows
    flag, kwargs = unpack_arg(diagonal_dominance)
    if flag:
        C = eliminate_diag_dom_nodes(A, C, **kwargs)

    # Compute the aggregation matrix AggOp (i.e., the nodal coarsening of A).
    # AggOp is a boolean matrix, where the sparsity pattern for the k-th column
    # denotes the fine-grid nodes agglomerated into k-th coarse-grid node.
    fn, kwargs = unpack_arg(aggregate[len(levels)-1])
    if fn == 'standard':
        AggOp, Cnodes = standard_aggregation(C, **kwargs)
    elif fn == 'naive':
        AggOp, Cnodes = naive_aggregation(C, **kwargs)
    elif fn == 'lloyd':
        AggOp, Cnodes = lloyd_aggregation(C, **kwargs)
    elif fn == 'pairwise':
        AggOp, Cnodes = pairwise_aggregation(A, **kwargs)
    elif fn == 'predefined':
        AggOp = kwargs['AggOp'].tocsr()
        Cnodes = kwargs['Cnodes']
    else:
        raise ValueError(f'Unrecognized aggregation method: {str(fn)}')

    # Improve near nullspace candidates by relaxing on A B = 0
    fn, kwargs = unpack_arg(improve_candidates[len(levels)-1])
    if fn is not None:
        b = np.zeros((A.shape[0], 1), dtype=A.dtype)
        B = relaxation_as_linear_operator((fn, kwargs), A, b) * B
        levels[-1].B = B
        if A.symmetry == 'nonsymmetric':
            BH = relaxation_as_linear_operator((fn, kwargs), AH, b) * BH
            levels[-1].BH = BH

    # Compute the tentative prolongator, T, which is a tentative interpolation
    # matrix from the coarse-grid to the fine-grid.  T exactly interpolates
    # B_fine[:, 0:get_blocksize(A)] = T B_coarse[:, 0:get_blocksize(A)].
    T, dummy = fit_candidates(AggOp, B[:, 0:get_blocksize(A)])
    del dummy
    if A.symmetry == 'nonsymmetric':
        TH, dummyH = fit_candidates(AggOp, BH[:, 0:get_blocksize(A)])
        del dummyH

    # Create necessary root node matrices
    Cpt_params = (True, get_Cpt_params(A, Cnodes, AggOp, T))
    T = scale_T(T, Cpt_params[1]['P_I'], Cpt_params[1]['I_F'])
    if A.symmetry == 'nonsymmetric':
        TH = scale_T(TH, Cpt_params[1]['P_I'], Cpt_params[1]['I_F'])

    # Set coarse grid near nullspace modes as injected fine grid near
    # null-space modes
    B = Cpt_params[1]['P_I'].T*levels[-1].B
    if A.symmetry == 'nonsymmetric':
        BH = Cpt_params[1]['P_I'].T*levels[-1].BH

    # Smooth the tentative prolongator, so that it's accuracy is greatly
    # improved for algebraically smooth error.
    fn, kwargs = unpack_arg(smooth[len(levels)-1])
    if fn == 'energy':
        P = energy_prolongation_smoother(A, T, C, B, levels[-1].B,
                                         Cpt_params=Cpt_params, **kwargs)
    elif fn is None:
        P = T
    else:
        raise ValueError(f'Unrecognized prolongation smoother method: {str(fn)}')

    # Compute the restriction matrix R, which interpolates from the fine-grid
    # to the coarse-grid.  If A is nonsymmetric, then R must be constructed
    # based on A.H.  Otherwise R = P.H or P.T.
    symmetry = A.symmetry
    if symmetry == 'hermitian':
        R = P.H
    elif symmetry == 'symmetric':
        R = P.T
    elif symmetry == 'nonsymmetric':
        fn, kwargs = unpack_arg(smooth[len(levels)-1])
        if fn == 'energy':
            R = energy_prolongation_smoother(AH, TH, C, BH, levels[-1].BH,
                                             Cpt_params=Cpt_params, **kwargs)
            R = R.H
        elif fn is None:
            R = T.H
        else:
            raise ValueError(f'Unrecognized prolongation smoother method: {str(fn)}')

    if keep:
        levels[-1].C = C                         # strength of connection matrix
        levels[-1].AggOp = AggOp                 # aggregation operator
        levels[-1].T = T                         # tentative prolongator
        levels[-1].Fpts = Cpt_params[1]['Fpts']  # Fpts
        levels[-1].P_I = Cpt_params[1]['P_I']    # Injection operator
        levels[-1].I_F = Cpt_params[1]['I_F']    # Identity on F-pts
        levels[-1].I_C = Cpt_params[1]['I_C']    # Identity on C-pts

    levels[-1].P = P                             # smoothed prolongator
    levels[-1].R = R                             # restriction operator
    levels[-1].Cpts = Cpt_params[1]['Cpts']      # Cpts (i.e., rootnodes)

    levels.append(MultilevelSolver.Level())
    A = R * A * P                                # Galerkin operator
    A.symmetry = symmetry
    levels[-1].A = A
    levels[-1].B = B                             # right near nullspace candidates

    if A.symmetry == 'nonsymmetric':
        levels[-1].BH = BH                       # left near nullspace candidates



"""Tentative prolongator."""


import numpy as np
from scipy.sparse import isspmatrix_csr, bsr_matrix
from pyamg import amg_core


def fit_candidates(AggOp, B, tol=1e-10):
    """Fit near-nullspace candidates to form the tentative prolongator.

    Parameters
    ----------
    AggOp : csr_matrix
        Describes the sparsity pattern of the tentative prolongator.
        Has dimension (#blocks, #aggregates)
    B : array
        The near-nullspace candidates stored in column-wise fashion.
        Has dimension (#blocks * blocksize, #candidates)
    tol : scalar
        Threshold for eliminating local basis functions.
        If after orthogonalization a local basis function Q[:, j] is small,
        i.e. ||Q[:, j]|| < tol, then Q[:, j] is set to zero.

    Returns
    -------
    (Q, R) : (bsr_matrix, array)
        The tentative prolongator Q is a sparse block matrix with dimensions
        (#blocks * blocksize, #aggregates * #candidates) formed by dense blocks
        of size (blocksize, #candidates).  The coarse level candidates are
        stored in R which has dimensions (#aggregates * #candidates,
        #candidates).

    See Also
    --------
    amg_core.fit_candidates

    Notes
    -----
        Assuming that each row of AggOp contains exactly one non-zero entry,
        i.e. all unknowns belong to an aggregate, then Q and R satisfy the
        relationship B = Q*R.  In other words, the near-nullspace candidates
        are represented exactly by the tentative prolongator.

        If AggOp contains rows with no non-zero entries, then the range of the
        tentative prolongator will not include those degrees of freedom. This
        situation is illustrated in the examples below.

    References
    ----------
    .. [1] Vanek, P. and Mandel, J. and Brezina, M.,
       "Algebraic Multigrid by Smoothed Aggregation for
       Second and Fourth Order Elliptic Problems",
       Computing, vol. 56, no. 3, pp. 179--196, 1996.
       http://citeseer.ist.psu.edu/vanek96algebraic.html


    Examples
    --------
    >>> from scipy.sparse import csr_matrix
    >>> from pyamg.aggregation.tentative import fit_candidates
    >>> # four nodes divided into two aggregates
    ... AggOp = csr_matrix( [[1, 0],
    ...                      [1, 0],
    ...                      [0, 1],
    ...                      [0, 1]] )
    >>> # B contains one candidate, the constant vector
    ... B = [[1],
    ...      [1],
    ...      [1],
    ...      [1]]
    >>> Q, R = fit_candidates(AggOp, B)
    >>> Q.toarray()
    array([[0.70710678, 0.        ],
           [0.70710678, 0.        ],
           [0.        , 0.70710678],
           [0.        , 0.70710678]])
    >>> R
    array([[1.41421356],
           [1.41421356]])
    >>> # Two candidates, the constant vector and a linear function
    ... B = [[1, 0],
    ...      [1, 1],
    ...      [1, 2],
    ...      [1, 3]]
    >>> Q, R = fit_candidates(AggOp, B)
    >>> Q.toarray()
    array([[ 0.70710678, -0.70710678,  0.        ,  0.        ],
           [ 0.70710678,  0.70710678,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  0.70710678, -0.70710678],
           [ 0.        ,  0.        ,  0.70710678,  0.70710678]])
    >>> R
    array([[1.41421356, 0.70710678],
           [0.        , 0.70710678],
           [1.41421356, 3.53553391],
           [0.        , 0.70710678]])
    >>> # aggregation excludes the third node
    ... AggOp = csr_matrix( [[1, 0],
    ...                      [1, 0],
    ...                      [0, 0],
    ...                      [0, 1]] )
    >>> B = [[1],
    ...      [1],
    ...      [1],
    ...      [1]]
    >>> Q, R = fit_candidates(AggOp, B)
    >>> Q.toarray()
    array([[0.70710678, 0.        ],
           [0.70710678, 0.        ],
           [0.        , 0.        ],
           [0.        , 1.        ]])
    >>> R
    array([[1.41421356],
           [1.        ]])

    """
    if not isspmatrix_csr(AggOp):
        raise TypeError('expected csr_matrix for argument AggOp')

    B = np.asarray(B)
    if B.dtype not in ['float32', 'float64', 'complex64', 'complex128']:
        B = np.asarray(B, dtype='float64')

    if len(B.shape) != 2:
        raise ValueError('expected 2d array for argument B')

    if B.shape[0] % AggOp.shape[0] != 0:
        raise ValueError(f'Dimensions of AggOp {AggOp.shape} and B {B.shape} '
                         'are incompatible')

    N_fine, N_coarse = AggOp.shape

    K1 = int(B.shape[0] / N_fine)  # dof per supernode (e.g. 3 for 3d vectors)
    K2 = B.shape[1]                # candidates

    # the first two dimensions of R and Qx are collapsed later
    R = np.empty((N_coarse, K2, K2), dtype=B.dtype)    # coarse candidates
    Qx = np.empty((AggOp.nnz, K1, K2), dtype=B.dtype)  # BSR data array

    AggOp_csc = AggOp.tocsc()

    fn = amg_core.fit_candidates
    fn(N_fine, N_coarse, K1, K2,
       AggOp_csc.indptr, AggOp_csc.indices, Qx.ravel(),
       B.ravel(), R.ravel(), tol)

    Q = bsr_matrix((Qx.swapaxes(1, 2).copy(), AggOp_csc.indices,
                    AggOp_csc.indptr), shape=(K2*N_coarse, K1*N_fine))
    Q = Q.T.tobsr()
    R = R.reshape(-1, K2)

    return Q, R



"""Aggregation methods."""


from warnings import warn
import numpy as np
from scipy import sparse
from .. import amg_core
from ..graph import lloyd_cluster
from ..strength import classical_strength_of_connection


def standard_aggregation(C):
    """Compute the sparsity pattern of the tentative prolongator.

    Parameters
    ----------
    C : csr_matrix
        strength of connection matrix

    Returns
    -------
    AggOp : csr_matrix
        aggregation operator which determines the sparsity pattern
        of the tentative prolongator
    Cpts : array
        array of Cpts, i.e., Cpts[i] = root node of aggregate i

    Examples
    --------
    >>> from scipy.sparse import csr_matrix
    >>> from pyamg.gallery import poisson
    >>> from pyamg.aggregation.aggregate import standard_aggregation
    >>> A = poisson((4,), format='csr')   # 1D mesh with 4 vertices
    >>> A.toarray()
    array([[ 2., -1.,  0.,  0.],
           [-1.,  2., -1.,  0.],
           [ 0., -1.,  2., -1.],
           [ 0.,  0., -1.,  2.]])
    >>> standard_aggregation(A)[0].toarray() # two aggregates
    array([[1, 0],
           [1, 0],
           [0, 1],
           [0, 1]], dtype=int8)
    >>> A = csr_matrix([[1,0,0],[0,1,1],[0,1,1]])
    >>> A.toarray()                      # first vertex is isolated
    array([[1, 0, 0],
           [0, 1, 1],
           [0, 1, 1]])
    >>> standard_aggregation(A)[0].toarray() # one aggregate
    array([[0],
           [1],
           [1]], dtype=int8)

    See Also
    --------
    amg_core.standard_aggregation

    """
    if not sparse.isspmatrix_csr(C):
        raise TypeError('expected csr_matrix')

    if C.shape[0] != C.shape[1]:
        raise ValueError('expected square matrix')

    index_type = C.indptr.dtype
    num_rows = C.shape[0]

    Tj = np.empty(num_rows, dtype=index_type)  # stores the aggregate #s
    Cpts = np.empty(num_rows, dtype=index_type)  # stores the Cpts

    fn = amg_core.standard_aggregation

    num_aggregates = fn(num_rows, C.indptr, C.indices, Tj, Cpts)
    Cpts = Cpts[:num_aggregates]

    # no nodes aggregated
    if num_aggregates == 0:
        # return all zero matrix and no Cpts
        return sparse.csr_matrix((num_rows, 1), dtype='int8'),\
            np.array([], dtype=index_type)

    shape = (num_rows, num_aggregates)

    # some nodes not aggregated
    if Tj.min() == -1:
        mask = Tj != -1
        row = np.arange(num_rows, dtype=index_type)[mask]
        col = Tj[mask]
        data = np.ones(len(col), dtype='int8')
        return sparse.coo_matrix((data, (row, col)), shape=shape).tocsr(), Cpts

    # all nodes aggregated
    Tp = np.arange(num_rows+1, dtype=index_type)
    Tx = np.ones(len(Tj), dtype='int8')
    return sparse.csr_matrix((Tx, Tj, Tp), shape=shape), Cpts


def naive_aggregation(C):
    """Compute the sparsity pattern of the tentative prolongator.

    Parameters
    ----------
    C : csr_matrix
        strength of connection matrix

    Returns
    -------
    AggOp : csr_matrix
        aggregation operator which determines the sparsity pattern
        of the tentative prolongator
    Cpts : array
        array of Cpts, i.e., Cpts[i] = root node of aggregate i

    Examples
    --------
    >>> from scipy.sparse import csr_matrix
    >>> from pyamg.gallery import poisson
    >>> from pyamg.aggregation.aggregate import naive_aggregation
    >>> A = poisson((4,), format='csr')   # 1D mesh with 4 vertices
    >>> A.toarray()
    array([[ 2., -1.,  0.,  0.],
           [-1.,  2., -1.,  0.],
           [ 0., -1.,  2., -1.],
           [ 0.,  0., -1.,  2.]])
    >>> naive_aggregation(A)[0].toarray() # two aggregates
    array([[1, 0],
           [1, 0],
           [0, 1],
           [0, 1]], dtype=int8)
    >>> A = csr_matrix([[1,0,0],[0,1,1],[0,1,1]])
    >>> A.toarray()                      # first vertex is isolated
    array([[1, 0, 0],
           [0, 1, 1],
           [0, 1, 1]])
    >>> naive_aggregation(A)[0].toarray() # two aggregates
    array([[1, 0],
           [0, 1],
           [0, 1]], dtype=int8)

    See Also
    --------
    amg_core.naive_aggregation

    Notes
    -----
    Differs from standard aggregation.  Each dof is considered.  If it has been
    aggregated, skip over.  Otherwise, put dof and any unaggregated neighbors
    in an aggregate.  Results in possibly much higher complexities than
    standard aggregation.

    """
    if not sparse.isspmatrix_csr(C):
        raise TypeError('expected csr_matrix')

    if C.shape[0] != C.shape[1]:
        raise ValueError('expected square matrix')

    index_type = C.indptr.dtype
    num_rows = C.shape[0]

    Tj = np.empty(num_rows, dtype=index_type)  # stores the aggregate #s
    Cpts = np.empty(num_rows, dtype=index_type)  # stores the Cpts

    fn = amg_core.naive_aggregation

    num_aggregates = fn(num_rows, C.indptr, C.indices, Tj, Cpts)
    Cpts = Cpts[:num_aggregates]
    Tj = Tj - 1

    if num_aggregates == 0:
        # all zero matrix
        return sparse.csr_matrix((num_rows, 1), dtype='int8'), Cpts

    shape = (num_rows, num_aggregates)
    # all nodes aggregated
    Tp = np.arange(num_rows+1, dtype=index_type)
    Tx = np.ones(len(Tj), dtype='int8')
    return sparse.csr_matrix((Tx, Tj, Tp), shape=shape), Cpts


def pairwise_aggregation(A, matchings=2, theta=0.25,
                         norm='min', compute_P=False):
    """Compute the sparsity pattern of the tentative prolongator.

    Parameters
    ----------
    A : csr_matrix or bsr_matrix
        level matrix
    matchings : int, default 2
        number of times to perform pairwise aggregation; each
        matching increases coarsening factor by about two.
    theta : float, default 0.25
        Strength tolerance used in computing classical SOC.
    norm : string, default 'min'
        Norm type used in computing classical SOC.
    compute_P : bool; default False
        Compute pairwise interpolation directly; if False, return
        integer aggregation matrix for smoothed_aggregation_solver.
        If True, return float interpolation P, converting to BSR
        form with identity of size bsize x bsize on each aggregate
        if A is BSR.

    Examples
    --------
    >>> from scipy.sparse import csr_matrix
    >>> from pyamg.gallery import poisson
    >>> from pyamg.aggregation.aggregate import pairwise_aggregation
    >>> A = poisson((4,), format='csr')   # 1D mesh with 4 vertices
    >>> A.toarray()
    array([[ 2., -1.,  0.,  0.],
           [-1.,  2., -1.,  0.],
           [ 0., -1.,  2., -1.],
           [ 0.,  0., -1.,  2.]])
    >>> pairwise_aggregation(A, matchings=1)[0].toarray() # two aggregates
    array([[1, 0],
           [1, 0],
           [0, 1],
           [0, 1]], dtype=int8)
    >>> pairwise_aggregation(A, matchings=2)[0].toarray() # one aggregate
    array([[1],
           [1],
           [1],
           [1]], dtype=int8)

    See Also
    --------
    amg_core.pairwise_aggregation

    References
    ----------
    [0] Notay, Y. (2010). An aggregation-based algebraic multigrid
    method. Electronic transactions on numerical analysis, 37(6),
    123-146.
    """
    # Get SOC matrix
    if not (sparse.isspmatrix_bsr(A) or sparse.isspmatrix_csr(A)):
        try:
            A = A.tocsr()
            warn('Implicit conversion of A to csr', sparse.SparseEfficiencyWarning)
        except BaseException as e:
            raise TypeError('Invalid matrix type, must be CSR or BSR.') from e

    index_type = A.indptr.dtype
    Ac = A      # Let Ac reference A for loop purposes
    T = None
    Cpts = None

    # Loop over the number of pairwise matchings to be done
    for i in range(0, matchings):

        # Compute SOC matrix for this matching
        if sparse.isspmatrix_bsr(A):
            C = classical_strength_of_connection(A=Ac, theta=theta, block=True, norm=norm)
        else:
            C = classical_strength_of_connection(A=Ac, theta=theta, block=False, norm=norm)

        # Form pairwise aggregation matrix
        num_rows = C.shape[0]
        Tj = np.empty(num_rows, dtype=index_type)  # stores the aggregate #s
        new_cpts = np.empty(num_rows, dtype=index_type)  # stores the new_cpts
        fn = amg_core.pairwise_aggregation
        num_aggregates = fn(num_rows, C.indptr, C.indices, C.data, Tj, new_cpts)
        if Cpts is None:
            Cpts = new_cpts[:num_aggregates]
        else:
            Cpts = Cpts[new_cpts[:num_aggregates]]
        Tj = Tj - 1

        # Construct sparse T
        if num_aggregates == 0:
            # all zero matrix
            T_temp = sparse.csr_matrix((num_rows, 1), dtype='int8')
            warn('No pairwise aggregates found, T = 0.')
        else:
            shape = (num_rows, num_aggregates)
            Tp = np.arange(num_rows+1, dtype=index_type)
            # If A is not BSR
            if not sparse.isspmatrix_bsr(A):
                Tx = np.ones(len(Tj), dtype='int8')
                T_temp = sparse.csr_matrix((Tx, Tj, Tp), shape=shape)
            else:
                shape = (shape[0]*A.blocksize[0], shape[1]*A.blocksize[1])
                Tx = np.array(len(Tj)*[np.identity(A.blocksize[0])], dtype='int8')
                T_temp = sparse.bsr_matrix((Tx, Tj, Tp), blocksize=A.blocksize, shape=shape)

        # Form aggregation matrix, need to make sure is CSR/BSR
        if i == 0:
            T = T_temp
        else:
            if sparse.isspmatrix_bsr(A):
                T = sparse.bsr_matrix(T * T_temp)
            else:
                T = sparse.csr_matrix(T * T_temp)

        # Break loop if zero aggregates were found
        if num_aggregates == 0:
            break

        # Form coarse grid operator for next matching
        if i < (matchings-1):
            if sparse.isspmatrix_csr(T_temp):
                Ac = T_temp.T.tocsr() * Ac * T_temp
            else:
                Ac = T_temp.T * Ac * T_temp

    # Convert T to dtype int if only used for aggregation
    if compute_P:
        T = T.astype(A.dtype, copy=False)

    return T, Cpts


def lloyd_aggregation(C, ratio=0.03, distance='unit', maxiter=10):
    """Aggregate nodes using Lloyd Clustering.

    Parameters
    ----------
    C : csr_matrix
        strength of connection matrix
    ratio : scalar
        Fraction of the nodes which will be seeds.
    distance : ['unit','abs','inv',None]
        Distance assigned to each edge of the graph G used in Lloyd clustering

        For each nonzero value C[i,j]:

        =======  ===========================
        'unit'   G[i,j] = 1
        'abs'    G[i,j] = abs(C[i,j])
        'inv'    G[i,j] = 1.0/abs(C[i,j])
        'same'   G[i,j] = C[i,j]
        'sub'    G[i,j] = C[i,j] - min(C)
        =======  ===========================

    maxiter : int
        Maximum number of iterations to perform

    Returns
    -------
    AggOp : csr_matrix
        aggregation operator which determines the sparsity pattern
        of the tentative prolongator
    seeds : array
        array of Cpts, i.e., Cpts[i] = root node of aggregate i

    See Also
    --------
    amg_core.standard_aggregation

    Examples
    --------
    >>> from scipy.sparse import csr_matrix
    >>> from pyamg.gallery import poisson
    >>> from pyamg.aggregation.aggregate import lloyd_aggregation
    >>> A = poisson((4,), format='csr')   # 1D mesh with 4 vertices
    >>> A.toarray()
    array([[ 2., -1.,  0.,  0.],
           [-1.,  2., -1.,  0.],
           [ 0., -1.,  2., -1.],
           [ 0.,  0., -1.,  2.]])
    >>> lloyd_aggregation(A)[0].toarray() # one aggregate
    array([[1],
           [1],
           [1],
           [1]], dtype=int8)
    >>> # more seeding for two aggregates
    >>> Agg = lloyd_aggregation(A,ratio=0.5)[0].toarray()

    """
    if ratio <= 0 or ratio > 1:
        raise ValueError('ratio must be > 0.0 and <= 1.0')

    if not (sparse.isspmatrix_csr(C) or sparse.isspmatrix_csc(C)):
        raise TypeError('expected csr_matrix or csc_matrix')

    if distance == 'unit':
        data = np.ones_like(C.data).astype(float)
    elif distance == 'abs':
        data = abs(C.data)
    elif distance == 'inv':
        data = 1.0/abs(C.data)
    elif distance == 'same':
        data = C.data
    elif distance == 'min':
        data = C.data - C.data.min()
    else:
        raise ValueError(f'Unrecognized value distance={distance}')

    if C.dtype == complex:
        data = np.real(data)

    assert data.min() >= 0

    G = C.__class__((data, C.indices, C.indptr), shape=C.shape)

    num_seeds = int(min(max(ratio * G.shape[0], 1), G.shape[0]))

    _, clusters, seeds = lloyd_cluster(G, num_seeds, maxiter=maxiter)

    row = (clusters >= 0).nonzero()[0]
    col = clusters[row]
    data = np.ones(len(row), dtype='int8')
    AggOp = sparse.coo_matrix((data, (row, col)),
                              shape=(G.shape[0], num_seeds)).tocsr()
    return AggOp, seeds


def balanced_lloyd_aggregation(C, num_clusters=None):
    """Aggregate nodes using Balanced Lloyd Clustering.

    Parameters
    ----------
    C : csr_matrix
        strength of connection matrix with positive weights
    num_clusters : int
        Number of seeds or clusters expected (default: C.shape[0] / 10)

    Returns
    -------
    AggOp : csr_matrix
        aggregation operator which determines the sparsity pattern
        of the tentative prolongator
    seeds : array
        array of Cpts, i.e., Cpts[i] = root node of aggregate i

    See Also
    --------
    amg_core.standard_aggregation

    Examples
    --------
    >>> import pyamg
    >>> from pyamg.aggregation.aggregate import balanced_lloyd_aggregation
    >>> data = pyamg.gallery.load_example('unit_square')
    >>> G = data['A']
    >>> xy = data['vertices'][:,:2]
    >>> G.data[:] = np.ones(len(G.data))

    >>> np.random.seed(787888)
    >>> AggOp, seeds = balanced_lloyd_aggregation(G)

    """
    if num_clusters is None:
        num_clusters = int(C.shape[0] / 10)

    if num_clusters < 1 or num_clusters > C.shape[0]:
        raise ValueError('num_clusters must be between 1 and n')

    if not (sparse.isspmatrix_csr(C) or sparse.isspmatrix_csc(C)):
        raise TypeError('expected csr_matrix or csc_matrix')

    if C.data.min() <= 0:
        raise ValueError('positive edge weights required')

    if C.dtype == complex:
        data = np.real(C.data)
    else:
        data = C.data

    G = C.__class__((data, C.indices, C.indptr), shape=C.shape)
    num_nodes = G.shape[0]

    seeds = np.random.permutation(num_nodes)[:num_clusters]
    seeds = seeds.astype(np.int32)
    mv = np.finfo(G.dtype).max
    d = mv * np.ones(num_nodes, dtype=G.dtype)
    d[seeds] = 0

    cm = -1 * np.ones(num_nodes, dtype=np.int32)
    cm[seeds] = seeds

    amg_core.lloyd_cluster_exact(num_nodes,
                                 G.indptr, G.indices, G.data,
                                 num_clusters,
                                 d, cm, seeds)

    col = cm
    row = np.arange(len(cm))
    data = np.ones(len(row), dtype=np.int32)
    AggOp = sparse.coo_matrix((data, (row, col)),
                              shape=(G.shape[0], num_clusters)).tocsr()
    return AggOp, seeds



"""Conjugate Gradient, Normal Error Krylov solver."""

import warnings
from warnings import warn
import numpy as np
from scipy import sparse
from scipy.sparse.linalg import aslinearoperator
from ..util.linalg import norm
from ..util import make_system


def cgne(A, b, x0=None, tol=1e-5, criteria='rr',
         maxiter=None, M=None,
         callback=None, residuals=None):
    """Conjugate Gradient, Normal Error algorithm.

    Applies CG to the normal equations, A A.H x = b. Left preconditioning
    is supported.  Note that unless A is well-conditioned, the use of
    CGNE is inadvisable

    Parameters
    ----------
    A : array, matrix, sparse matrix, LinearOperator
        n x n, linear system to solve
    b : array, matrix
        right hand side, shape is (n,) or (n,1)
    x0 : array, matrix
        initial guess, default is a vector of zeros
    tol : float
        Tolerance for stopping criteria
    criteria : string
        Stopping criteria, let r=r_k, x=x_k
        'rr':        ||r||       < tol ||b||
        'rr+':       ||r||       < tol (||b|| + ||A||_F ||x||)
        'MrMr':      ||M r||     < tol ||M b||
        'rMr':       <r, Mr>^1/2 < tol
        if ||b||=0, then set ||b||=1 for these tests.
    maxiter : int
        maximum number of iterations allowed
    M : array, matrix, sparse matrix, LinearOperator
        n x n, inverted preconditioner, i.e. solve M A A.H x = M b.
    callback : function
        User-supplied function is called after each iteration as
        callback(xk), where xk is the current solution vector
    residuals : list
        residual history in the 2-norm, including the initial residual

    Returns
    -------
    (xk, info)
    xk : an updated guess after k iterations to the solution of Ax = b
    info : halting status

            ==  =======================================
            0   successful exit
            >0  convergence to tolerance not achieved,
                return iteration count instead.
            <0  numerical breakdown, or illegal input
            ==  =======================================

    Notes
    -----
    The LinearOperator class is in scipy.sparse.linalg.
    Use this class if you prefer to define A or M as a mat-vec routine
    as opposed to explicitly constructing the matrix.

    Examples
    --------
    >>> from pyamg.krylov import cgne
    >>> from pyamg.util.linalg import norm
    >>> import numpy as np
    >>> from pyamg.gallery import poisson
    >>> A = poisson((10,10))
    >>> b = np.ones((A.shape[0],))
    >>> (x,flag) = cgne(A,b, maxiter=2, tol=1e-8)
    >>> print(f'{norm(b - A*x):.6}')
    46.1547

    References
    ----------
    .. [1] Yousef Saad, "Iterative Methods for Sparse Linear Systems,
       Second Edition", SIAM, pp. 276-7, 2003
       http://www-users.cs.umn.edu/~saad/books.html

    """
    # Store the conjugate transpose explicitly as it will be used much later on
    if sparse.isspmatrix(A):
        AH = A.H
    else:
        # avoid doing this since A may be a different sparse type
        AH = aslinearoperator(np.asarray(A).conj().T)

    # Convert inputs to linear system, with error checking
    A, M, x, b, postprocess = make_system(A, M, x0, b)
    n = A.shape[0]

    # Ensure that warnings are always reissued from this function
    warnings.filterwarnings('always', module='pyamg.krylov._cgne')

    # How often should r be recomputed
    recompute_r = 8

    # Check iteration numbers. CGNE suffers from loss of orthogonality quite
    # easily, so we arbitrarily let the method go up to 130% over the
    # theoretically necessary limit of maxiter=n
    if maxiter is None:
        maxiter = int(np.ceil(1.3*n)) + 2
    elif maxiter < 1:
        raise ValueError('Number of iterations must be positive')
    elif maxiter > (1.3*n):
        warn('maximum allowed inner iterations (maxiter) are the 130% times'
             'the number of dofs')
        maxiter = int(np.ceil(1.3*n)) + 2

    # Prep for method
    r = b - A @ x
    normr = norm(r)

    # Apply preconditioner and calculate initial search direction
    z = M @ r
    p = AH @ z
    old_zr = np.inner(z.conjugate(), r)

    if residuals is not None:
        residuals[:] = [normr]  # initial residual

    # Check initial guess if b != 0,
    normb = norm(b)
    if normb == 0.0:
        normb = 1.0  # reset so that tol is unscaled

    # set the stopping criteria (see the docstring)
    if criteria == 'rr':
        rtol = tol * normb
    elif criteria == 'rr+':
        if sparse.issparse(A.A):
            normA = norm(A.A.data)
        elif isinstance(A.A, np.ndarray):
            normA = norm(np.ravel(A.A))
        else:
            raise ValueError('Unable to use ||A||_F with the current matrix format.')
        rtol = tol * (normA * np.linalg.norm(x) + normb)
    elif criteria == 'MrMr':
        normr = norm(z)
        normMb = norm(M @ b)
        rtol = tol * normMb
    elif criteria == 'rMr':
        normr = np.sqrt(old_zr)
        rtol = tol
    else:
        raise ValueError('Invalid stopping criteria.')

    if normr < rtol:
        return (postprocess(x), 0)

    # Begin CGNE

    it = 0

    while True:                                   # Step number in Saad's pseudocode

        # alpha = (z_j, r_j) / (p_j, p_j)
        alpha = old_zr / np.inner(p.conjugate(), p)

        # x_{j+1} = x_j + alpha*p_j
        x += alpha * p

        # r_{j+1} = r_j - alpha*w_j,   where w_j = A*p_j
        if np.mod(it, recompute_r) and it > 0:
            r -= alpha * (A @ p)
        else:
            r = b - A @ x

        # z_{j+1} = M*r_{j+1}
        z = M @ r

        # beta = (z_{j+1}, r_{j+1}) / (z_j, r_j)
        new_zr = np.inner(z.conjugate(), r)
        beta = new_zr / old_zr
        old_zr = new_zr

        # p_{j+1} = A.H*z_{j+1} + beta*p_j
        p *= beta
        p += AH @ z

        it += 1
        normr = np.linalg.norm(r)

        if residuals is not None:
            residuals.append(normr)

        if callback is not None:
            callback(x)

        # set the stopping criteria (see the docstring)
        if criteria == 'rr':
            rtol = tol * normb
        elif criteria == 'rr+':
            rtol = tol * (normA * np.linalg.norm(x) + normb)
        elif criteria == 'MrMr':
            normr = norm(z)
            rtol = tol * normMb
        elif criteria == 'rMr':
            normr = np.sqrt(new_zr)
            rtol = tol

        if normr < rtol:
            return (postprocess(x), 0)

        if it == maxiter:
            return (postprocess(x), it)
